# -*- coding: utf-8 -*-
"""
Dependencies: flask, tornado

SeeAlso:
    routes.turk_identification
"""
import logging
from wbia.control import controller_inject
from wbia.algo.hots import pipeline
from flask import url_for, request, current_app  # NOQA
from os.path import join, dirname, abspath, exists
import cv2
import numpy as np  # NOQA
import utool as ut
from wbia.web import appfuncs as appf
from wbia import constants as const
import traceback
import requests
import six
from datetime import datetime

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')

CLASS_INJECT_KEY, register_ibs_method = controller_inject.make_ibs_register_decorator(
    __name__
)
register_api = controller_inject.get_wbia_flask_api(__name__)
register_route = controller_inject.get_wbia_flask_route(__name__)


GRAPH_CLIENT_PEEK = 100


@register_ibs_method
@register_api('/api/query/annot/rowid/', methods=['GET'])
def get_recognition_query_aids(ibs, is_known, species=None):
    """
    DEPCIRATE

    RESTful:
        Method: GET
        URL:    /api/query/annot/rowid/
    """
    qaid_list = ibs.get_valid_aids(is_known=is_known, species=species)
    return qaid_list


@register_ibs_method
@register_api('/api/query/chip/dict/simple/', methods=['GET'])
def query_chips_simple_dict(ibs, *args, **kwargs):
    r"""
    Runs query_chips, but returns a json compatible dictionary

    Args:
        same as query_chips

    RESTful:
        Method: GET
        URL:    /api/query/chip/dict/simple/

    SeeAlso:
        query_chips

    CommandLine:
        python -m wbia.web.apis_query --test-query_chips_simple_dict:0
        python -m wbia.web.apis_query --test-query_chips_simple_dict:1

        python -m wbia.web.apis_query --test-query_chips_simple_dict:0 --humpbacks

    Example:
        >>> # xdoctest: +REQUIRES(--web-tests)
        >>> from wbia.control.IBEISControl import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> #qaid = ibs.get_valid_aids()[0:3]
        >>> qaids = ibs.get_valid_aids()
        >>> daids = ibs.get_valid_aids()
        >>> dict_list = ibs.query_chips_simple_dict(qaids, daids)
        >>> qgids = ibs.get_annot_image_rowids(qaids)
        >>> qnids = ibs.get_annot_name_rowids(qaids)
        >>> for dict_, qgid, qnid in list(zip(dict_list, qgids, qnids)):
        >>>     dict_['qgid'] = qgid
        >>>     dict_['qnid'] = qnid
        >>>     dict_['dgid_list'] = ibs.get_annot_image_rowids(dict_['daid_list'])
        >>>     dict_['dnid_list'] = ibs.get_annot_name_rowids(dict_['daid_list'])
        >>>     dict_['dgname_list'] = ibs.get_image_gnames(dict_['dgid_list'])
        >>>     dict_['qgname'] = ibs.get_image_gnames(dict_['qgid'])
        >>> result  = ut.repr2(dict_list, nl=2, precision=2, hack_liststr=True)
        >>> result = result.replace('u\'', '"').replace('\'', '"')
        >>> print(result)

    Example:
        >>> # xdoctest: +SKIP
        >>> # FIXME failing-test (04-Aug-2020) This test hangs when running together with the test above
        >>> from wbia.control.IBEISControl import *  # NOQA
        >>> import wbia
        >>> # Start up the web instance
        >>> with wbia.opendb_bg_web(db='testdb1', managed=True) as web_ibs:
        ...     cmdict_list = web_ibs.send_wbia_request('/api/query/chip/dict/simple/', type_='get', qaid_list=[1], daid_list=[1, 2, 3])
        >>> print(cmdict_list)
        >>> assert 'score_list' in cmdict_list[0]

    """
    kwargs['return_cm_simple_dict'] = True
    return ibs.query_chips(*args, **kwargs)


@register_ibs_method
@register_api('/api/query/chip/dict/', methods=['GET'])
def query_chips_dict(ibs, *args, **kwargs):
    """
    Runs query_chips, but returns a json compatible dictionary

    RESTful:
        Method: GET
        URL:    /api/query/chip/dict/
    """
    kwargs['return_cm_dict'] = True
    return ibs.query_chips(*args, **kwargs)


@register_ibs_method
@register_api('/api/review/query/graph/', methods=['POST'])
def process_graph_match_html(ibs, **kwargs):
    """
    RESTful:
        Method: POST
        URL:    /api/review/query/graph/
    """

    def sanitize(state):
        state = state.strip().lower()
        state = ''.join(state.split())
        return state

    import uuid

    map_dict = {
        'sameanimal': const.EVIDENCE_DECISION.INT_TO_CODE[
            const.EVIDENCE_DECISION.POSITIVE
        ],
        'differentanimals': const.EVIDENCE_DECISION.INT_TO_CODE[
            const.EVIDENCE_DECISION.NEGATIVE
        ],
        'cannottell': const.EVIDENCE_DECISION.INT_TO_CODE[
            const.EVIDENCE_DECISION.INCOMPARABLE
        ],
        'unreviewed': const.EVIDENCE_DECISION.INT_TO_CODE[
            const.EVIDENCE_DECISION.UNREVIEWED
        ],
        'unknown': const.EVIDENCE_DECISION.INT_TO_CODE[const.EVIDENCE_DECISION.UNKNOWN],
        'photobomb': 'photobomb',
        'scenerymatch': 'scenerymatch',
        'excludetop': 'excludetop',
        'excludebottom': 'excludebottom',
        'excludeleft': 'excludeleft',
        'excluderight': 'excluderight',
    }
    annot_uuid_1 = uuid.UUID(request.form['identification-annot-uuid-1'])
    annot_uuid_2 = uuid.UUID(request.form['identification-annot-uuid-2'])
    state = request.form.get('identification-submit', '')
    state = sanitize(state)
    state = map_dict[state]
    tag_list = []
    if state in ['photobomb', 'scenerymatch']:
        tag_list.append(state)
        state = const.EVIDENCE_DECISION.NEGATIVE
    assert state in map_dict.values(), 'matching state is unrecognized'
    # Get checbox tags
    checbox_tag_list = ['photobomb', 'scenerymatch']
    for checbox_tag in checbox_tag_list:
        checkbox_name = 'ia-%s-value' % (checbox_tag)
        if checkbox_name in request.form:
            tag_list.append(checbox_tag)
    tag_list = sorted(set(tag_list))
    confidence_default = const.CONFIDENCE.INT_TO_CODE[const.CONFIDENCE.UNKNOWN]
    confidence = request.form.get('ia-turk-confidence', confidence_default)
    if confidence not in const.CONFIDENCE.CODE_TO_INT.keys():
        confidence = confidence_default
    if len(tag_list) == 0:
        tag_str = ''
    else:
        tag_str = ';'.join(tag_list)
    user_times = {
        'server_time_start': request.form.get('server_time_start', None),
        'client_time_start': request.form.get('client_time_start', None),
        'client_time_end': request.form.get('client_time_end', None),
    }
    return (
        annot_uuid_1,
        annot_uuid_2,
        state,
        tag_str,
        'web-api',
        confidence,
        user_times,
    )


def ensure_review_image(
    ibs,
    aid,
    cm,
    qreq_,
    view_orientation='vertical',
    draw_matches=True,
    draw_heatmask=False,
    verbose=False,
):
    r""" "
    Create the review image for a pair of annotations

    CommandLine:
        python -m wbia.web.apis_query ensure_review_image --show

    Example:
        >>> # SCRIPT
        >>> from wbia.web.apis_query import *  # NOQA
        >>> import wbia
        >>> cm, qreq_ = wbia.testdata_cm('PZ_MTEST', a='default:dindex=0:10,qindex=0:1')
        >>> ibs = qreq_.ibs
        >>> aid = cm.get_top_aids()[0]
        >>> tt = ut.tic('make image')
        >>> image, _ = ensure_review_image(ibs, aid, cm, qreq_)
        >>> ut.toc(tt)
        >>> ut.quit_if_noshow()
        >>> print('image.shape = %r' % (image.shape,))
        >>> print('image.dtype = %r' % (image.dtype,))
        >>> ut.print_object_size(image)
        >>> import wbia.plottool as pt
        >>> pt.imshow(image)
        >>> ut.show_if_requested()
    """
    from wbia.gui import id_review_api

    # Get thumb path
    match_thumb_path = ibs.get_match_thumbdir()
    match_thumb_filename = id_review_api.get_match_thumb_fname(
        cm,
        aid,
        qreq_,
        view_orientation=view_orientation,
        draw_matches=draw_matches,
        draw_heatmask=draw_heatmask,
    )
    match_thumb_filepath = join(match_thumb_path, match_thumb_filename)
    if verbose:
        logger.info('Checking: %r' % (match_thumb_filepath,))

    if exists(match_thumb_filepath):
        image = cv2.imread(match_thumb_filepath)
    else:
        render_config = {
            'dpi': 300,
            'overlay': True,
            'draw_fmatches': True,
            'draw_fmatch': draw_matches,
            'show_matches': draw_matches,
            'show_ell': draw_matches,
            'show_lines': draw_matches,
            'show_ori': False,
            'heatmask': draw_heatmask,
            'vert': view_orientation == 'vertical',
            'show_aidstr': False,
            'show_name': False,
            'show_exemplar': False,
            'show_num_gt': False,
            'show_timedelta': False,
            'show_name_rank': False,
            'show_score': False,
            'show_annot_score': False,
            'show_name_score': False,
            'draw_lbl': False,
            'draw_border': False,
            'white_background': True,
        }

        if hasattr(qreq_, 'render_single_result'):
            image = qreq_.render_single_result(cm, aid, **render_config)
        else:
            image = cm.render_single_annotmatch(qreq_, aid, **render_config)
        # image = vt.crop_out_imgfill(image, fillval=(255, 255, 255), thresh=64)
        cv2.imwrite(match_thumb_filepath, image)
    return image, match_thumb_filepath


@register_api(
    '/api/review/query/graph/alias/', methods=['POST'], __api_plural_check__=False
)
def review_graph_match_html_alias(*args, **kwargs):
    review_graph_match_html(*args, **kwargs)


@register_api('/api/review/query/graph/', methods=['GET'])
def review_graph_match_html(
    ibs,
    review_pair,
    cm_dict,
    query_config_dict,
    _internal_state,
    callback_url,
    callback_method='POST',
    view_orientation='vertical',
    include_jquery=False,
):
    r"""
    Args:
        ibs (wbia.IBEISController):  image analysis api
        review_pair (dict): pair of annot uuids
        cm_dict (dict):
        query_config_dict (dict):
        _internal_state (?):
        callback_url (str):
        callback_method (unicode): (default = u'POST')
        view_orientation (unicode): (default = u'vertical')
        include_jquery (bool): (default = False)

    CommandLine:
        python -m wbia.web.apis_query review_graph_match_html --show

        wbia --web
        python -m wbia.web.apis_query review_graph_match_html --show --domain=localhost

    Example:
        >>> # xdoctest: +REQUIRES(--web-tests)
        >>> from wbia.web.apis_query import *  # NOQA
        >>> import wbia
        >>> web_ibs = wbia.opendb_bg_web('testdb1')  # , domain='http://52.33.105.88')
        >>> aids = web_ibs.send_wbia_request('/api/annot/', 'get')[0:2]
        >>> uuid_list = web_ibs.send_wbia_request('/api/annot/uuid/', type_='get', aid_list=aids)
        >>> quuid_list = uuid_list[0:1]
        >>> duuid_list = uuid_list
        >>> query_config_dict = {
        >>>    # 'pipeline_root' : 'BC_DTW'
        >>> }
        >>> data = dict(
        >>>     query_annot_uuid_list=quuid_list, database_annot_uuid_list=duuid_list,
        >>>     query_config_dict=query_config_dict,
        >>> )
        >>> jobid = web_ibs.send_wbia_request('/api/engine/query/graph/', **data)
        >>> print('jobid = %r' % (jobid,))
        >>> status_response = web_ibs.wait_for_results(jobid)
        >>> result_response = web_ibs.read_engine_results(jobid)
        >>> inference_result = result_response['json_result']
        >>> print('inference_result = %r' % (inference_result,))
        >>> auuid2_cm = inference_result['cm_dict']
        >>> quuid = quuid_list[0]
        >>> class_dict = auuid2_cm[str(quuid)]
        >>> # Get information in frontend
        >>> #ibs = wbia.opendb('testdb1')
        >>> #cm = match_obj = wbia.ChipMatch.from_dict(class_dict, ibs=ibs)
        >>> #match_obj.print_rawinfostr()
        >>> # Make the dictionary a bit more managable
        >>> #match_obj.compress_top_feature_matches(num=2)
        >>> #class_dict = match_obj.to_dict(ibs=ibs)
        >>> cm_dict = class_dict
        >>> # Package for review
        >>> review_pair = {'annot_uuid_1': quuid, 'annot_uuid_2': duuid_list[1]}
        >>> callback_method = u'POST'
        >>> view_orientation = u'vertical'
        >>> include_jquery = False
        >>> kw = dict(
        >>>     review_pair=review_pair,
        >>>     cm_dict=cm_dict,
        >>>     query_config_dict=query_config_dict,
        >>>     _internal_state=None,
        >>>     callback_url = None,
        >>> )
        >>> html_str = web_ibs.send_wbia_request('/api/review/query/graph/', type_='get', **kw)
        >>> web_ibs.terminate2()
        >>> ut.quit_if_noshow()
        >>> import wbia.plottool as pt
        >>> ut.render_html(html_str)
        >>> ut.show_if_requested()

    Example2:
        >>> # DISABLE_DOCTEST
        >>> # This starts off using web to get information, but finishes the rest in python
        >>> from wbia.web.apis_query import *  # NOQA
        >>> import wbia
        >>> ut.exec_funckw(review_graph_match_html, globals())
        >>> web_ibs = wbia.opendb_bg_web('testdb1')  # , domain='http://52.33.105.88')
        >>> aids = web_ibs.send_wbia_request('/api/annot/', 'get')[0:2]
        >>> uuid_list = web_ibs.send_wbia_request('/api/annot/uuid/', type_='get', aid_list=aids)
        >>> quuid_list = uuid_list[0:1]
        >>> duuid_list = uuid_list
        >>> query_config_dict = {
        >>>    # 'pipeline_root' : 'BC_DTW'
        >>> }
        >>> data = dict(
        >>>     query_annot_uuid_list=quuid_list, database_annot_uuid_list=duuid_list,
        >>>     query_config_dict=query_config_dict,
        >>> )
        >>> jobid = web_ibs.send_wbia_request('/api/engine/query/graph/', **data)
        >>> status_response = web_ibs.wait_for_results(jobid)
        >>> result_response = web_ibs.read_engine_results(jobid)
        >>> web_ibs.terminate2()
        >>> # NOW WORK IN THE FRONTEND
        >>> inference_result = result_response['json_result']
        >>> auuid2_cm = inference_result['cm_dict']
        >>> quuid = quuid_list[0]
        >>> class_dict = auuid2_cm[str(quuid)]
        >>> # Get information in frontend
        >>> ibs = wbia.opendb('testdb1')
        >>> cm = wbia.ChipMatch.from_dict(class_dict, ibs=ibs)
        >>> cm.print_rawinfostr()
        >>> # Make the dictionary a bit more managable
        >>> cm.compress_top_feature_matches(num=1)
        >>> cm.print_rawinfostr()
        >>> class_dict = cm.to_dict(ibs=ibs)
        >>> cm_dict = class_dict
        >>> # Package for review ( CANT CALL DIRECTLY BECAUSE OF OUT OF CONTEXT )
        >>> review_pair = {'annot_uuid_1': quuid, 'annot_uuid_2': duuid_list[1]}
        >>> x = review_graph_match_html(ibs, review_pair, cm_dict,
        >>>                             query_config_dict, _internal_state=None,
        >>>                             callback_url=None)
        >>> ut.quit_if_noshow()
        >>> import wbia.plottool as pt
        >>> ut.render_html(html_str)
        >>> ut.show_if_requested()
    """
    from wbia.algo.hots import chip_match

    # from wbia.algo.hots.query_request import QueryRequest

    proot = query_config_dict.get('pipeline_root', 'vsmany')
    proot = query_config_dict.get('proot', proot)
    if proot.lower() in (
        'bc_dtw',
        'oc_wdtw',
        'curvrankdorsal',
        'curvrankfinfindrhybriddorsal',
        'curvrankfluke',
        'deepsense',
        'finfindr',
        'kaggle7',
        'kaggleseven',
        'pie',
    ):
        cls = chip_match.AnnotMatch  # ibs.depc_annot.requestclass_dict['BC_DTW']
    else:
        cls = chip_match.ChipMatch

    view_orientation = view_orientation.lower()
    if view_orientation not in ['vertical', 'horizontal']:
        view_orientation = 'horizontal'

    # unpack info
    try:
        annot_uuid_1 = review_pair['annot_uuid_1']
        annot_uuid_2 = review_pair['annot_uuid_2']
    except Exception:
        # ??? HACK
        # FIXME:
        logger.info('[!!!!] review_pair = %r' % (review_pair,))
        review_pair = review_pair[0]
        annot_uuid_1 = review_pair['annot_uuid_1']
        annot_uuid_2 = review_pair['annot_uuid_2']

    ibs.web_check_uuids(qannot_uuid_list=[annot_uuid_1], dannot_uuid_list=[annot_uuid_2])

    aid_1 = ibs.get_annot_aids_from_uuid(annot_uuid_1)
    aid_2 = ibs.get_annot_aids_from_uuid(annot_uuid_2)

    cm = cls.from_dict(cm_dict, ibs=ibs)
    qreq_ = ibs.new_query_request([aid_1], [aid_2], cfgdict=query_config_dict)

    # Get score
    idx = cm.daid2_idx[aid_2]
    match_score = cm.name_score_list[idx]
    # match_score = cm.aid2_score[aid_2]

    try:
        image_matches, _ = ensure_review_image(
            ibs, aid_2, cm, qreq_, view_orientation=view_orientation
        )
    except KeyError:
        image_matches = np.zeros((100, 100, 3), dtype=np.uint8)
        traceback.print_exc()
    try:
        image_clean, _ = ensure_review_image(
            ibs, aid_2, cm, qreq_, view_orientation=view_orientation, draw_matches=False
        )
    except KeyError:
        image_clean = np.zeros((100, 100, 3), dtype=np.uint8)
        traceback.print_exc()

    image_matches_src = appf.embed_image_html(image_matches)
    image_clean_src = appf.embed_image_html(image_clean)

    confidence_dict = const.CONFIDENCE.NICE_TO_CODE
    confidence_nice_list = confidence_dict.keys()
    confidence_text_list = confidence_dict.values()
    confidence_selected_list = [
        confidence_text == 'unspecified' for confidence_text in confidence_text_list
    ]
    confidence_list = list(
        zip(confidence_nice_list, confidence_text_list, confidence_selected_list)
    )

    if False:
        from wbia.web import apis_query

        root_path = dirname(abspath(apis_query.__file__))
    else:
        root_path = dirname(abspath(__file__))
    css_file_list = [
        ['css', 'style.css'],
        ['include', 'bootstrap', 'css', 'bootstrap.css'],
    ]
    json_file_list = [
        ['javascript', 'script.js'],
        ['include', 'bootstrap', 'js', 'bootstrap.js'],
    ]

    if include_jquery:
        json_file_list = [['javascript', 'jquery.min.js']] + json_file_list

    EMBEDDED_CSS = ''
    EMBEDDED_JAVASCRIPT = ''

    css_template_fmtstr = '<style type="text/css" ia-dependency="css">%s</style>\n'
    json_template_fmtstr = (
        '<script type="text/javascript" ia-dependency="javascript">%s</script>\n'
    )
    for css_file in css_file_list:
        css_filepath_list = [root_path, 'static'] + css_file
        with open(join(*css_filepath_list)) as css_file:
            EMBEDDED_CSS += css_template_fmtstr % (css_file.read(),)

    for json_file in json_file_list:
        json_filepath_list = [root_path, 'static'] + json_file
        with open(join(*json_filepath_list)) as json_file:
            EMBEDDED_JAVASCRIPT += json_template_fmtstr % (json_file.read(),)

    annot_uuid_1 = str(annot_uuid_1)
    annot_uuid_2 = str(annot_uuid_2)

    embedded = dict(globals(), **locals())
    return appf.template('turk', 'identification_insert', **embedded)


@register_route('/test/review/query/chip/', methods=['GET'])
def review_query_chips_test(**kwargs):
    """
    CommandLine:
        python -m wbia.web.apis_query review_query_chips_test --show

    Example:
        >>> # SCRIPT
        >>> import wbia
        >>> web_ibs = wbia.opendb_bg_web(
        >>>     browser=True, url_suffix='/test/review/query/chip/?__format__=true')
    """
    ibs = current_app.ibs

    # the old block curvature dtw
    if 'use_bc_dtw' in request.args:
        query_config_dict = {'pipeline_root': 'BC_DTW'}
    # the new oriented curvature dtw
    elif 'use_oc_wdtw' in request.args:
        query_config_dict = {'pipeline_root': 'OC_WDTW'}
    elif 'use_curvrank_dorsal' in request.args:
        query_config_dict = {'pipeline_root': 'CurvRankDorsal'}
    elif 'use_curvrank_finfindr_hybrid_dorsal' in request.args:
        query_config_dict = {'pipeline_root': 'CurvRankFinfindrHybridDorsal'}
    elif 'use_curvrank_fluke' in request.args:
        query_config_dict = {'pipeline_root': 'CurvRankFluke'}
    elif 'use_deepsense' in request.args:
        query_config_dict = {'pipeline_root': 'Deepsense'}
    elif 'use_finfindr' in request.args:
        query_config_dict = {'pipeline_root': 'Finfindr'}
    elif 'use_kaggle7' in request.args or 'use_kaggleseven' in request.args:
        query_config_dict = {'pipeline_root': 'KaggleSeven'}
    elif 'use_pie' in request.args:
        query_config_dict = {'pipeline_root': 'Pie'}
    else:
        query_config_dict = {}
    result_dict = ibs.query_chips_test(query_config_dict=query_config_dict)

    review_pair = result_dict['inference_dict']['annot_pair_dict']['review_pair_list'][0]
    annot_uuid_key = str(review_pair['annot_uuid_key'])
    cm_dict = result_dict['cm_dict'][annot_uuid_key]
    query_config_dict = result_dict['query_config_dict']
    _internal_state = result_dict['inference_dict']['_internal_state']
    callback_url = request.args.get('callback_url', url_for('process_graph_match_html'))
    callback_method = request.args.get('callback_method', 'POST')
    # view_orientation = request.args.get('view_orientation', 'vertical')
    view_orientation = request.args.get('view_orientation', 'horizontal')

    template_html = review_graph_match_html(
        ibs,
        review_pair,
        cm_dict,
        query_config_dict,
        _internal_state,
        callback_url,
        callback_method,
        view_orientation,
        include_jquery=True,
    )
    template_html = """
        <script src="http://code.jquery.com/jquery-2.2.1.min.js" ia-dependency="javascript"></script>
        %s
    """ % (
        template_html,
    )
    return template_html
    return 'done'


@register_ibs_method
@register_api('/api/review/query/chip/best/', methods=['GET'])
def review_query_chips_best(ibs, aid, **kwargs):
    # from wbia.algo.hots import chip_match

    # query_config_dict = {}

    # Compile test data
    qaid_list = [aid]
    daid_list = ibs.get_valid_aids()
    result_dict = ibs.query_chips_graph(qaid_list, daid_list, **kwargs)

    uuid = list(result_dict['cm_dict'].keys())[0]
    result = result_dict['cm_dict'][uuid]

    scores = result['name_score_list']
    names = result['unique_name_list']
    best_index = np.argmax(scores)

    best_name = names[best_index]

    return best_name

    # review_pair = result_dict['inference_dict']['annot_pair_dict']['review_pair_list'][0]

    # annot_uuid_key = str(review_pair['annot_uuid_key'])
    # cm_dict = result_dict['cm_dict'][annot_uuid_key]
    # query_config_dict = result_dict['query_config_dict']

    # proot = query_config_dict.get('pipeline_root', 'vsmany')
    # proot = query_config_dict.get('proot', proot)

    # view_orientation = 'horizontal'

    # # unpack info
    # try:
    #     annot_uuid_1 = review_pair['annot_uuid_1']
    #     annot_uuid_2 = review_pair['annot_uuid_2']
    # except Exception:
    #     #??? HACK
    #     # FIXME:
    #     logger.info('[!!!!] review_pair = %r' % (review_pair,))
    #     review_pair = review_pair[0]
    #     annot_uuid_1 = review_pair['annot_uuid_1']
    #     annot_uuid_2 = review_pair['annot_uuid_2']

    # ibs.web_check_uuids(qannot_uuid_list=[annot_uuid_1],
    #                     dannot_uuid_list=[annot_uuid_2])

    # aid_1 = ibs.get_annot_aids_from_uuid(annot_uuid_1)
    # aid_2 = ibs.get_annot_aids_from_uuid(annot_uuid_2)

    # cm = chip_match.ChipMatch.from_dict(cm_dict, ibs=ibs)
    # qreq_ = ibs.new_query_request([aid_1], [aid_2],
    #                               cfgdict=query_config_dict)

    # # Get score
    # # idx = cm.daid2_idx[aid_2]
    # # match_score = cm.name_score_list[idx]
    # match_score = cm.aid2_score[aid_2]

    # try:
    #     image_matches, _ = ensure_review_image(ibs, aid_2, cm, qreq_,
    #                                         view_orientation=view_orientation)
    # except KeyError:
    #     image_matches = np.zeros((100, 100, 3), dtype=np.uint8)
    #     traceback.print_exc()
    # try:
    #     image_clean, _ = ensure_review_image(ibs, aid_2, cm, qreq_,
    #                                       view_orientation=view_orientation,
    #                                       draw_matches=False)
    # except KeyError:
    #     image_clean = np.zeros((100, 100, 3), dtype=np.uint8)
    #     traceback.print_exc()

    # image_matches_src = appf.embed_image_html(image_matches)
    # image_clean_src = appf.embed_image_html(image_clean)

    # return_dict = {
    #     'match': image_matches_src,
    #     'clean': image_clean_src,
    #     'score': match_score,
    # }
    # return return_dict


@register_ibs_method
@register_api('/test/query/chip/', methods=['GET'])
def query_chips_test(ibs, aid=None, limited=False, census_annotations=True, **kwargs):
    """
    CommandLine:
        python -m wbia.web.apis_query query_chips_test

    Example:
        >>> # SLOW_DOCTEST
        >>> # xdoctest: +SKIP
        >>> from wbia.control.IBEISControl import *  # NOQA
        >>> import wbia
        >>> qreq_ = wbia.testdata_qreq_(defaultdb='testdb1')
        >>> ibs = qreq_.ibs
        >>> result_dict = ibs.query_chips_test()
        >>> print(result_dict)
    """
    from random import shuffle  # NOQA

    # Compile test data
    aid_list = ibs.get_valid_aids()
    if aid is None:
        shuffle(aid_list)
        qaid_list = aid_list[:1]
    else:
        qaid_list = [aid]
    daid_list = aid_list

    if limited:
        daid_list = daid_list[-4:]

    if census_annotations:
        flag_list = ibs.get_annot_canonical(daid_list)
        daid_list = ut.compress(daid_list, flag_list)

    result_dict = ibs.query_chips_graph(qaid_list, daid_list, **kwargs)
    return result_dict


@register_ibs_method
@register_api('/api/query/graph/complete/', methods=['GET', 'POST'])
def query_chips_graph_complete(ibs, aid_list, query_config_dict={}, k=5, **kwargs):
    import theano  # NOQA

    cm_list, qreq_ = ibs.query_chips(
        qaid_list=aid_list,
        daid_list=aid_list,
        cfgdict=query_config_dict,
        return_request=True,
    )

    result_dict = {}
    for cm in cm_list:
        key = str(ibs.get_annot_uuids(cm.qaid))
        id_list = ibs.get_annot_uuids(cm.daid_list)
        score_list = cm.annot_score_list
        value_list = sorted(zip(score_list, id_list), reverse=True)
        k_ = min(k, len(value_list))
        value_list = value_list[:k_]
        result = {str(id_): score for score, id_ in value_list}
        result_dict[key] = result

    return result_dict


@register_ibs_method
@register_api('/api/query/graph/', methods=['GET', 'POST'])
def query_chips_graph(
    ibs,
    qaid_list,
    daid_list,
    user_feedback=None,
    query_config_dict={},
    echo_query_params=True,
    cache_images=True,
    n=16,
    view_orientation='horizontal',
    return_summary=True,
    **kwargs,
):
    from wbia.unstable.orig_graph_iden import OrigAnnotInference
    import theano  # NOQA
    import uuid

    def convert_to_uuid(nid):
        try:
            text = ibs.get_name_texts(nid)
            uuid_ = uuid.UUID(text)
        except ValueError:
            uuid_ = nid
        return uuid_

    proot = query_config_dict.get('pipeline_root', 'vsmany')
    proot = query_config_dict.get('proot', proot)
    logger.info('query_config_dict = %r' % (query_config_dict,))

    curvrank_daily_tag = query_config_dict.get('curvrank_daily_tag', None)
    if curvrank_daily_tag is not None:
        if len(curvrank_daily_tag) > 144:
            curvrank_daily_tag_ = ut.hashstr27(curvrank_daily_tag)
            curvrank_daily_tag_ = 'wbia-shortened-%s' % (curvrank_daily_tag_,)
            logger.info('[WARNING] curvrank_daily_tag too long (Probably an old job)')
            logger.info('[WARNING] Original: %r' % (curvrank_daily_tag,))
            logger.info('[WARNING] Shortened: %r' % (curvrank_daily_tag_,))
            query_config_dict['curvrank_daily_tag'] = curvrank_daily_tag_

    num_qaids = len(qaid_list)
    num_daids = len(daid_list)

    valid_aid_list = ibs.get_valid_aids()
    qaid_list = list(set(qaid_list) & set(valid_aid_list))
    daid_list = list(set(daid_list) & set(valid_aid_list))

    num_qaids_ = len(qaid_list)
    num_daids_ = len(daid_list)

    if num_qaids != num_qaids_:
        logger.info('len(qaid_list)  = %d' % (num_qaids,))
        logger.info('len(qaid_list_) = %d' % (num_qaids_,))

    if num_daids != num_daids_:
        logger.info('len(daid_list)  = %d' % (num_daids,))
        logger.info('len(daid_list_) = %d' % (num_daids_,))

    if num_qaids_ == 0:
        raise ValueError('There are 0 valid query aids, %d were provided' % (num_qaids,))

    if num_daids_ == 0:
        raise ValueError(
            'There are 0 valid database aids, %d were provided' % (num_daids,)
        )

    cm_list, qreq_ = ibs.query_chips(
        qaid_list=qaid_list,
        daid_list=daid_list,
        cfgdict=query_config_dict,
        return_request=True,
    )

    annot_inference = OrigAnnotInference(qreq_, cm_list, user_feedback)
    inference_dict = annot_inference.make_annot_inference_dict()

    cache_dir = join(ibs.cachedir, 'query_match')
    ut.ensuredir(cache_dir)

    cache_path = None
    reference = None

    if cache_images:
        reference = ut.hashstr27(qreq_.get_cfgstr())
        cache_path = join(cache_dir, 'qreq_cfgstr_%s' % (reference,))
        ut.ensuredir(cache_path)

    cm_dict = {}
    for cm in cm_list:
        quuid = ibs.get_annot_uuids(cm.qaid)
        cm_key = str(quuid)

        if cache_images:
            args = (quuid,)
            qannot_cache_filepath = join(cache_path, 'qannot_uuid_%s' % args)
            ut.ensuredir(qannot_cache_filepath)

            assert cache_path is not None

            score_list = cm.score_list
            daid_list_ = cm.daid_list

            # Get best names
            logger.info('Visualizing name matches')
            DEEPSENSE_NUM_TO_VISUALIZE_PER_NAME = 3

            unique_nids = cm.unique_nids
            name_score_list = cm.name_score_list

            zipped = sorted(zip(name_score_list, unique_nids), reverse=True)
            n_ = min(n, len(zipped))
            zipped = zipped[:n_]
            logger.info('Top %d names: %r' % (n_, zipped))
            dnid_set = set(ut.take_column(zipped, 1))
            aids_list = ibs.get_name_aids(dnid_set)

            daid_set = []
            for nid, aid_list in zip(dnid_set, aids_list):
                # Assume that the highest AIDs are the newest sightings, visualize the most recent X sightings per name
                aid_list = sorted(aid_list, reverse=True)
                num_per_name = (
                    len(aid_list)
                    if DEEPSENSE_NUM_TO_VISUALIZE_PER_NAME is None
                    else DEEPSENSE_NUM_TO_VISUALIZE_PER_NAME
                )
                limit = min(len(aid_list), num_per_name)
                args = (
                    nid,
                    len(aid_list),
                    limit,
                )
                logger.info('\tFiltering NID %d from %d -> %d' % args)
                aid_list = aid_list[:limit]
                daid_set += aid_list

            args = (
                len(daid_set),
                daid_set,
            )
            logger.info('Found %d candidate name aids: %r' % args)
            daid_set = set(daid_set)
            name_daid_set = daid_set & set(daid_list_)  # Filter by supplied query daids

            args = (
                len(name_daid_set),
                len(dnid_set),
                name_daid_set,
            )
            logger.info('Found %d overlapping aids for %d nids: %r' % args)

            args = (
                len(name_daid_set),
                len(dnid_set),
            )
            logger.info('Visualizing %d annotations for best %d names' % args)

            # Get best annotations
            logger.info('Visualizing annotation matches')
            zipped = sorted(zip(score_list, daid_list_), reverse=True)
            n_ = min(n, len(zipped))
            zipped = zipped[:n_]
            logger.info('Top %d annots: %r' % (n_, zipped))
            annot_daid_set = set(ut.take_column(zipped, 1))

            # Combine names and annotations
            daid_set = list(name_daid_set) + list(annot_daid_set)
            daid_set = list(set(daid_set))
            logger.info('Visualizing %d annots: %r' % (len(daid_set), daid_set))

            extern_flag_list = []
            for daid in daid_list_:
                extern_flag = daid in daid_set

                if extern_flag:
                    logger.info('Rendering match images to disk for daid=%d' % (daid,))
                    duuid = ibs.get_annot_uuids(daid)

                    args = (duuid,)
                    dannot_cache_filepath = join(
                        qannot_cache_filepath, 'dannot_uuid_%s' % args
                    )
                    ut.ensuredir(dannot_cache_filepath)

                    cache_filepath_fmtstr = join(
                        dannot_cache_filepath, 'version_%s_orient_%s.png'
                    )

                    try:
                        _, filepath_matches = ensure_review_image(
                            ibs,
                            daid,
                            cm,
                            qreq_,
                            view_orientation=view_orientation,
                            draw_matches=True,
                            draw_heatmask=False,
                        )
                    except Exception:
                        filepath_matches = None
                        extern_flag = 'error'
                    try:
                        _, filepath_heatmask = ensure_review_image(
                            ibs,
                            daid,
                            cm,
                            qreq_,
                            view_orientation=view_orientation,
                            draw_matches=False,
                            draw_heatmask=True,
                        )
                    except Exception:
                        filepath_heatmask = None
                        extern_flag = 'error'
                    try:
                        _, filepath_clean = ensure_review_image(
                            ibs,
                            daid,
                            cm,
                            qreq_,
                            view_orientation=view_orientation,
                            draw_matches=False,
                            draw_heatmask=False,
                        )
                    except Exception:
                        filepath_clean = None
                        extern_flag = 'error'

                    if filepath_matches is not None:
                        args = (
                            'matches',
                            view_orientation,
                        )
                        cache_filepath = cache_filepath_fmtstr % args
                        ut.symlink(filepath_matches, cache_filepath, overwrite=True)

                    if filepath_heatmask is not None:
                        args = (
                            'heatmask',
                            view_orientation,
                        )
                        cache_filepath = cache_filepath_fmtstr % args
                        ut.symlink(filepath_heatmask, cache_filepath, overwrite=True)

                    if filepath_clean is not None:
                        args = (
                            'clean',
                            view_orientation,
                        )
                        cache_filepath = cache_filepath_fmtstr % args
                        ut.symlink(filepath_clean, cache_filepath, overwrite=True)

                extern_flag_list.append(extern_flag)
        else:
            extern_flag_list = None

        cm_dict[cm_key] = {
            # 'qaid'                    : cm.qaid,
            'qannot_uuid': ibs.get_annot_uuids(cm.qaid),
            # 'qnid'                    : cm.qnid,
            'qname_uuid': convert_to_uuid(cm.qnid),
            'qname': ibs.get_name_texts(cm.qnid),
            # 'daid_list'               : cm.daid_list,
            'dannot_uuid_list': ibs.get_annot_uuids(cm.daid_list),
            # 'dnid_list'               : cm.dnid_list,
            'dname_uuid_list': [convert_to_uuid(nid) for nid in cm.dnid_list],
            # FIXME: use qreq_ state not wbia state
            'dname_list': ibs.get_name_texts(cm.dnid_list),
            'score_list': cm.score_list,
            'annot_score_list': cm.annot_score_list,
            'fm_list': cm.fm_list if hasattr(cm, 'fm_list') else None,
            'fsv_list': cm.fsv_list if hasattr(cm, 'fsv_list') else None,
            # Non-corresponding lists to above
            # 'unique_nids'             : cm.unique_nids,
            'unique_name_uuid_list': [convert_to_uuid(nid) for nid in cm.unique_nids],
            # FIXME: use qreq_ state not wbia state
            'unique_name_list': ibs.get_name_texts(cm.unique_nids),
            'name_score_list': cm.name_score_list,
            # Placeholders for the reinitialization of the ChipMatch object
            'fk_list': None,
            'H_list': None,
            'fsv_col_lbls': None,
            'filtnorm_aids': None,
            'filtnorm_fxs': None,
            'dannot_extern_reference': reference,
            'dannot_extern_list': extern_flag_list,
        }

    result_dict = {
        'cm_dict': cm_dict,
        'inference_dict': inference_dict,
    }

    if echo_query_params:
        result_dict['query_annot_uuid_list'] = ibs.get_annot_uuids(qaid_list)
        result_dict['database_annot_uuid_list'] = ibs.get_annot_uuids(daid_list)
        result_dict['query_config_dict'] = query_config_dict

    if return_summary:
        summary_list = [
            ('summary_annot', cm.annot_score_list),
            ('summary_name', cm.name_score_list),
        ]
        for summary_key, summary_score_list in summary_list:
            value_list = sorted(list(zip(summary_score_list, cm.daid_list)), reverse=True)
            n_ = min(len(value_list), n)
            value_list = value_list[:n_]

            dscore_list = ut.take_column(value_list, 0)
            daid_list_ = ut.take_column(value_list, 1)
            duuid_list = ibs.get_annot_uuids(daid_list_)
            dnid_list = ibs.get_annot_nids(daid_list_)
            species_list = ibs.get_annot_species(daid_list_)
            viewpoint_list = ibs.get_annot_viewpoints(daid_list_)
            values_list = list(
                zip(
                    dscore_list,
                    duuid_list,
                    daid_list_,
                    dnid_list,
                    species_list,
                    viewpoint_list,
                )
            )
            key_list = ['score', 'duuid', 'daid', 'dnid', 'species', 'viewpoint']
            result_dict[summary_key] = [
                dict(zip(key_list, value_list_)) for value_list_ in values_list
            ]

    return result_dict


@register_route(
    '/api/query/graph/match/thumb/',
    methods=['GET'],
    __route_prefix_check__=False,
    __route_authenticate__=False,
)
def query_chips_graph_match_thumb(
    extern_reference, query_annot_uuid, database_annot_uuid, version
):
    from PIL import Image  # NOQA
    import vtool as vt
    from io import BytesIO
    from six.moves import cStringIO as StringIO
    from flask import send_file

    ibs = current_app.ibs
    args = (
        extern_reference,
        query_annot_uuid,
        database_annot_uuid,
        version,
    )

    cache_dir = join(ibs.cachedir, 'query_match')

    reference_path = join(cache_dir, 'qreq_cfgstr_%s' % (extern_reference,))
    if not exists(reference_path):
        message = 'dannot_extern_reference is unknown'
        raise controller_inject.WebMatchThumbException(*args, message=message)

    qannot_path = join(reference_path, 'qannot_uuid_%s' % (query_annot_uuid,))
    if not exists(qannot_path):
        message = 'query_annot_uuid is unknown for the given reference %s' % (
            extern_reference,
        )
        raise controller_inject.WebMatchThumbException(*args, message=message)

    dannot_path = join(qannot_path, 'dannot_uuid_%s' % (database_annot_uuid,))
    if not exists(dannot_path):
        message = (
            'database_annot_uuid is unknown for the given reference %s and query_annot_uuid %s'
            % (extern_reference, query_annot_uuid)
        )
        raise controller_inject.WebMatchThumbException(*args, message=message)

    version = version.lower()

    alias_dict = {
        None: 'clean',
        'match': 'matches',
        'mask': 'heatmask',
    }
    for alias in alias_dict:
        version = alias_dict.get(version, version)

    assert version in ['clean', 'matches', 'heatmask']

    version_path = join(dannot_path, 'version_%s_orient_horizontal.png' % (version,))
    if not exists(version_path):
        message = (
            'match thumb version is unknown for the given reference %s, query_annot_uuid %s, database_annot_uuid %s'
            % (extern_reference, query_annot_uuid, database_annot_uuid)
        )
        raise controller_inject.WebMatchThumbException(*args, message=message)

    # Load image
    image = vt.imread(version_path, orient='auto')
    image = image[:, :, ::-1]

    # Encode image
    image_pil = Image.fromarray(image)
    if six.PY2:
        img_io = StringIO()
    else:
        img_io = BytesIO()
    image_pil.save(img_io, 'JPEG', quality=100)
    img_io.seek(0)
    return send_file(img_io, mimetype='image/jpeg')


@register_ibs_method
@register_api('/api/query/chip/', methods=['GET'])
def query_chips(
    ibs,
    qaid_list=None,
    daid_list=None,
    cfgdict=None,
    use_cache=None,
    use_bigcache=None,
    qreq_=None,
    return_request=False,
    verbose=pipeline.VERB_PIPELINE,
    save_qcache=None,
    prog_hook=None,
    return_cm_dict=False,
    return_cm_simple_dict=False,
):
    r"""
    Submits a query request to the hotspotter recognition pipeline. Returns
    a list of QueryResult objects.

    Args:
        qaid_list (list): a list of annotation ids to be submitted as
            queries
        daid_list (list): a list of annotation ids used as the database
            that will be searched
        cfgdict (dict): dictionary of configuration options used to create
            a new QueryRequest if not already specified
        use_cache (bool): turns on/off chip match cache (default: True)
        use_bigcache (bool): turns one/off chunked chip match cache (default:
            True)
        qreq_ (QueryRequest): optional, a QueryRequest object that
            overrides all previous settings
        return_request (bool): returns the request which will be created if
            one is not already specified
        verbose (bool): default=False, turns on verbose printing

    Returns:
        list: a list of ChipMatch objects containing the matching
            annotations, scores, and feature matches

    Returns(2):
        tuple: (cm_list, qreq_) - a list of query results and optionally the
            QueryRequest object used

    RESTful:
        Method: PUT
        URL:    /api/query/chip/

    CommandLine:
        python -m wbia.web.apis_query --test-query_chips

        # Test speed of single query
        python -m wbia --tf IBEISController.query_chips --db PZ_Master1 \
            -a default:qindex=0:1,dindex=0:500 --nocache-hs

        python -m wbia --tf IBEISController.query_chips --db PZ_Master1 \
            -a default:qindex=0:1,dindex=0:3000 --nocache-hs

        python -m wbia.web.apis_query --test-query_chips:1 --show
        python -m wbia.web.apis_query --test-query_chips:2 --show

    Example:
        >>> # SLOW_DOCTEST
        >>> # xdoctest: +SKIP
        >>> from wbia.control.IBEISControl import *  # NOQA
        >>> import wbia
        >>> qreq_ = wbia.testdata_qreq_()
        >>> ibs = qreq_.ibs
        >>> cm_list = qreq_.execute()
        >>> cm = cm_list[0]
        >>> ut.quit_if_noshow()
        >>> cm.ishow_analysis(qreq_)
        >>> ut.show_if_requested()

    Example:
        >>> # SLOW_DOCTEST
        >>> # xdoctest: +SKIP
        >>> import wbia
        >>> from wbia.control.IBEISControl import *  # NOQA
        >>> qaid_list = [1]
        >>> daid_list = [1, 2, 3, 4, 5]
        >>> ibs = wbia.opendb_test(db='testdb1')
        >>> qreq_ = ibs.new_query_request(qaid_list, daid_list)
        >>> cm = ibs.query_chips(qaid_list, daid_list, use_cache=False, qreq_=qreq_)[0]
        >>> ut.quit_if_noshow()
        >>> cm.ishow_analysis(qreq_)
        >>> ut.show_if_requested()

    Example1:
        >>> # SLOW_DOCTEST
        >>> # xdoctest: +SKIP
        >>> import wbia
        >>> from wbia.control.IBEISControl import *  # NOQA
        >>> qaid_list = [1]
        >>> daid_list = [1, 2, 3, 4, 5]
        >>> ibs = wbia.opendb_test(db='testdb1')
        >>> cfgdict = {'pipeline_root':'BC_DTW'}
        >>> qreq_ = ibs.new_query_request(qaid_list, daid_list, cfgdict=cfgdict, verbose=True)
        >>> cm = ibs.query_chips(qreq_=qreq_)[0]
        >>> ut.quit_if_noshow()
        >>> cm.ishow_analysis(qreq_)
        >>> ut.show_if_requested()
    """
    # The qaid and daid objects are allowed to be None if qreq_ is
    # specified
    if qreq_ is None:
        assert qaid_list is not None, 'Need to specify qaids'
        assert daid_list is not None, 'Need to specify daids'
        qaid_list, was_scalar = ut.wrap_iterable(qaid_list)
        if daid_list is None:
            daid_list = ibs.get_valid_aids()
        qreq_ = ibs.new_query_request(
            qaid_list, daid_list, cfgdict=cfgdict, verbose=verbose
        )
    else:
        assert qaid_list is None, 'do not specify qreq and qaids'
        assert daid_list is None, 'do not specify qreq and daids'
        was_scalar = False
    cm_list = qreq_.execute()
    assert isinstance(cm_list, list), 'Chip matches were not returned as a list'

    # Convert to cm_list
    if return_cm_simple_dict:
        for cm in cm_list:
            cm.qauuid = ibs.get_annot_uuids(cm.qaid)
            cm.dauuid_list = ibs.get_annot_uuids(cm.daid_list)
        keys = ['qaid', 'daid_list', 'score_list', 'qauuid', 'dauuid_list']
        cm_list = [ut.dict_subset(cm.to_dict(), keys) for cm in cm_list]
    elif return_cm_dict:
        cm_list = [cm.to_dict() for cm in cm_list]

    if was_scalar:
        # hack for scalar input
        assert len(cm_list) == 1
        cm_list = cm_list[0]

    if return_request:
        return cm_list, qreq_
    else:
        return cm_list


##########################################################################################


@register_ibs_method
def get_graph_client_query_chips_graph_v2(ibs, graph_uuid):
    graph_client = current_app.GRAPH_CLIENT_DICT.get(graph_uuid, None)
    # We could be redirecting to a newer graph_client
    graph_uuid_chain = [graph_uuid]
    while isinstance(graph_client, six.string_types):
        graph_uuid_chain.append(graph_client)
        graph_client = current_app.GRAPH_CLIENT_DICT.get(graph_client, None)
    if graph_client is None:
        raise controller_inject.WebUnknownUUIDException(['graph_uuid'], [graph_uuid])
    return graph_client, graph_uuid_chain


def ensure_review_image_v2(
    ibs,
    match,
    draw_matches=False,
    draw_heatmask=False,
    view_orientation='vertical',
    overlay=True,
):
    import wbia.plottool as pt

    render_config = {
        'overlay': overlay,
        'show_ell': draw_matches,
        'show_lines': draw_matches,
        'show_ori': False,
        'heatmask': draw_heatmask,
        'vert': view_orientation == 'vertical',
    }
    with pt.RenderingContext(dpi=150) as ctx:
        match.show(**render_config)
    image = ctx.image
    return image


def query_graph_v2_callback(graph_client, callback_type):
    from wbia.web.graph_server import ut_to_json_encode

    assert callback_type in ['review', 'finished']
    callback_tuple = graph_client.callbacks.get(callback_type, None)
    if callback_tuple is not None:
        callback_url, callback_method = callback_tuple
        if callback_url is not None:
            callback_method = callback_method.lower()
            data_dict = ut_to_json_encode({'graph_uuid': graph_client.graph_uuid})
            if callback_method == 'post':
                requests.post(callback_url, data=data_dict)
            elif callback_method == 'get':
                requests.get(callback_url, params=data_dict)
            elif callback_method == 'put':
                requests.put(callback_url, data=data_dict)
            elif callback_method == 'delete':
                requests.delete(callback_url, data=data_dict)
            else:
                raise KeyError('Unsupported HTTP callback method')


@register_ibs_method
@register_api('/api/query/graph/v2/', methods=['POST'])
def query_chips_graph_v2(
    ibs,
    annot_uuid_list=None,
    query_config_dict={},
    review_callback_url=None,
    review_callback_method='POST',
    finished_callback_url=None,
    finished_callback_method='POST',
    creation_imageset_rowid_list=None,
    **kwargs,
):
    """
    CommandLine:
        python -m wbia.web.apis_query --test-query_chips_graph_v2:0

        python -m wbia reset_mtest_graph

        python -m wbia --db PZ_MTEST --web --browser --url=/turk/identification/hardcase/
        python -m wbia --db PZ_MTEST --web --browser --url=/turk/identification/graph/

    Example:
        >>> # xdoctest: +REQUIRES(--web-tests)
        >>> from wbia.web.apis_query import *
        >>> import wbia
        >>> # Open local instance
        >>> ibs = wbia.opendb('PZ_MTEST')
        >>> uuid_list = ibs.annots().uuids[0:10]
        >>> # Start up the web instance
        >>> web_ibs = wbia.opendb_bg_web(db='PZ_MTEST', web=True, browser=False)
        >>> data = dict(annot_uuid_list=uuid_list)
        >>> resp = web_ibs.send_wbia_request('/api/query/graph/v2/', **data)
        >>> print('resp = %r' % (resp,))
        >>> #cmdict_list = json_dict['response']
        >>> #assert 'score_list' in cmdict_list[0]

    Example:
        >>> # DEBUG_SCRIPT
        >>> from wbia.web.apis_query import *
        >>> # Hack a flask context
        >>> current_app = ut.DynStruct()
        >>> current_app.GRAPH_CLIENT_DICT = {}
        >>> old = query_chips_graph_v2.__globals__.get('current_app', None)
        >>> query_chips_graph_v2.__globals__['current_app'] = current_app
        >>> import wbia
        >>> ibs = wbia.opendb('PZ_MTEST')
        >>> #ut.exec_funckw(query_chips_graph_v2, globals())
        >>> # Run function in main process
        >>> query_chips_graph_v2(ibs)
        >>> # Reset context
        >>> query_chips_graph_v2.__globals__['current_app'] = old
    """
    from wbia.web.graph_server import GraphClient

    logger.info('[apis_query] Creating GraphClient')

    if annot_uuid_list is None:
        annot_uuid_list = ibs.get_annot_uuids(ibs.get_valid_aids())

    ibs.web_check_uuids(qannot_uuid_list=annot_uuid_list)
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)

    # FILTER FOR GGR2
    if False:
        aid_list = ibs.check_ggr_valid_aids(aid_list, **kwargs)

    graph_uuid = ut.hashable_to_uuid(sorted(aid_list))
    if graph_uuid not in current_app.GRAPH_CLIENT_DICT:
        for graph_uuid_ in current_app.GRAPH_CLIENT_DICT:
            graph_client_ = current_app.GRAPH_CLIENT_DICT[graph_uuid_]
            aid_list_ = graph_client_.aids
            assert aid_list_ is not None
            overlap_aid_set = set(aid_list_) & set(aid_list)
            if len(overlap_aid_set) > 0:
                overlap_aid_list = list(overlap_aid_set)
                overlap_annot_uuid_list = ibs.get_annot_uuids(overlap_aid_list)
                raise controller_inject.WebUnavailableUUIDException(
                    overlap_annot_uuid_list, graph_uuid_
                )

        callback_dict = {
            'review': (review_callback_url, review_callback_method),
            'finished': (finished_callback_url, finished_callback_method),
        }
        graph_client = GraphClient(graph_uuid, callbacks=callback_dict, autoinit=True)

        if creation_imageset_rowid_list is not None:
            graph_client.imagesets = creation_imageset_rowid_list
        graph_client.aids = aid_list

        config = {
            'manual.n_peek': GRAPH_CLIENT_PEEK,
            'manual.autosave': True,
            'redun.pos': 2,
            'redun.neg': 2,
            'algo.quickstart': False,
        }
        config.update(query_config_dict)
        logger.info('[apis_query] graph_client.config = {}'.format(ut.repr3(config)))
        graph_client.config = config

        # Ensure no race-conditions
        current_app.GRAPH_CLIENT_DICT[graph_uuid] = graph_client

        # Start (create the Graph Inference object)
        payload = {
            'action': 'start',
            'dbdir': ibs.dbdir,
            'aids': graph_client.aids,
            'config': graph_client.config,
        }
        future = graph_client.post(payload)
        future.result()  # Guarantee that this has happened before calling refresh

        f2 = graph_client.post({'action': 'latest_logs'})
        f2.graph_client = graph_client
        f2.add_done_callback(query_graph_v2_latest_logs)

        # Start (create the Graph Inference object)
        payload = {
            'action': 'get_feat_extractor',
        }
        future = graph_client.post(payload)
        graph_client.extr = future.result()

        # Start main loop
        future = graph_client.post({'action': 'continue_review'})
        future.graph_client = graph_client
        future.add_done_callback(query_graph_v2_on_request_review)

        f2 = graph_client.post({'action': 'latest_logs'})
        f2.graph_client = graph_client
        f2.add_done_callback(query_graph_v2_latest_logs)

    return graph_uuid


@register_ibs_method
def review_graph_match_config_v2(
    ibs, graph_uuid, aid1=None, aid2=None, view_orientation='vertical', view_version=1
):
    from wbia.algo.verif import pairfeat
    from flask import session

    EDGES_KEY = '_EDGES_'
    EDGES_MAX = 10

    user_id = controller_inject.get_user()
    graph_client, _ = ibs.get_graph_client_query_chips_graph_v2(graph_uuid)

    if aid1 is not None and aid2 is not None:
        previous_edge_list = None
        if aid1 > aid2:
            aid1, aid2 = aid2, aid1
        edge = (aid1, aid2)
        data = graph_client.check(edge)
        if data is None:
            data = (
                edge,
                np.nan,
                {},
            )
    else:
        if EDGES_KEY not in session:
            session[EDGES_KEY] = []
        previous_edge_list = session[EDGES_KEY]
        logger.info(
            'Using previous_edge_list\n\tUser: %s\n\tList: %r'
            % (user_id, previous_edge_list)
        )

        data = graph_client.sample(
            previous_edge_list=previous_edge_list, max_previous_edges=EDGES_MAX
        )
        if data is None:
            raise controller_inject.WebReviewNotReadyException(graph_uuid)

    edge, priority, data_dict = data
    edge = [
        int(edge[0]),
        int(edge[1]),
    ]
    aid_1, aid_2 = edge
    annot_uuid_1 = str(ibs.get_annot_uuids(aid_1))
    annot_uuid_2 = str(ibs.get_annot_uuids(aid_2))

    if previous_edge_list is not None:
        previous_edge_list.append(edge)
        if len(previous_edge_list) > EDGES_MAX:
            cutoff = int(-1.0 * EDGES_MAX)
            previous_edge_list = previous_edge_list[cutoff:]
        session[EDGES_KEY] = previous_edge_list
        logger.info(
            'Updating previous_edge_list\n\tUser: %s\n\tList: %r'
            % (user_id, previous_edge_list)
        )

    args = (
        edge,
        priority,
    )
    logger.info('Sampled edge %r with priority %0.02f' % args)
    logger.info('Data: ' + ut.repr4(data_dict))

    feat_extract_config = {
        'match_config': (
            {} if graph_client.extr is None else graph_client.extr.match_config
        )
    }
    extr = pairfeat.PairwiseFeatureExtractor(ibs, config=feat_extract_config)

    match = extr._exec_pairwise_match([edge])[0]

    image_clean = ensure_review_image_v2(
        ibs, match, view_orientation=view_orientation, overlay=False
    )
    # image_matches = ensure_review_image_v2(ibs, match, draw_matches=True,
    #                                        view_orientation=view_orientation)

    logger.info('Using View Version: %r' % (view_version,))
    if view_version == 1:
        image_heatmask = ensure_review_image_v2(
            ibs, match, draw_heatmask=True, view_orientation=view_orientation
        )
    else:
        image_heatmask = ensure_review_image_v2(
            ibs, match, draw_matches=True, view_orientation=view_orientation
        )

    image_clean_src = appf.embed_image_html(image_clean)
    # image_matches_src = appf.embed_image_html(image_matches)
    image_heatmask_src = appf.embed_image_html(image_heatmask)

    now = datetime.utcnow()
    server_time_start = float(now.strftime('%s.%f'))

    return (
        edge,
        priority,
        data_dict,
        aid_1,
        aid_2,
        annot_uuid_1,
        annot_uuid_2,
        image_clean_src,
        image_heatmask_src,
        image_heatmask_src,
        server_time_start,
    )


@register_api('/api/review/query/graph/v2/', methods=['GET'])
def review_graph_match_html_v2(
    ibs,
    graph_uuid,
    callback_url=None,
    callback_method='POST',
    view_orientation='vertical',
    view_version=1,
    include_jquery=False,
):
    values = ibs.review_graph_match_config_v2(
        graph_uuid, view_orientation=view_orientation, view_version=view_version
    )

    (
        edge,
        priority,
        data_dict,
        aid1,
        aid2,
        annot_uuid_1,
        annot_uuid_2,
        image_clean_src,
        image_matches_src,
        image_heatmask_src,
        server_time_start,
    ) = values

    confidence_dict = const.CONFIDENCE.NICE_TO_CODE
    confidence_nice_list = confidence_dict.keys()
    confidence_text_list = confidence_dict.values()
    confidence_selected_list = [
        confidence_text == 'unspecified' for confidence_text in confidence_text_list
    ]
    confidence_list = list(
        zip(confidence_nice_list, confidence_text_list, confidence_selected_list)
    )

    if False:
        from wbia.web import apis_query

        root_path = dirname(abspath(apis_query.__file__))
    else:
        root_path = dirname(abspath(__file__))
    css_file_list = [
        ['css', 'style.css'],
        ['include', 'bootstrap', 'css', 'bootstrap.css'],
    ]
    json_file_list = [
        ['javascript', 'script.js'],
        ['include', 'bootstrap', 'js', 'bootstrap.js'],
    ]

    if include_jquery:
        json_file_list = [['javascript', 'jquery.min.js']] + json_file_list

    EMBEDDED_CSS = ''
    EMBEDDED_JAVASCRIPT = ''

    css_template_fmtstr = '<style type="text/css" ia-dependency="css">%s</style>\n'
    json_template_fmtstr = (
        '<script type="text/javascript" ia-dependency="javascript">%s</script>\n'
    )
    for css_file in css_file_list:
        css_filepath_list = [root_path, 'static'] + css_file
        with open(join(*css_filepath_list)) as css_file:
            EMBEDDED_CSS += css_template_fmtstr % (css_file.read(),)

    for json_file in json_file_list:
        json_filepath_list = [root_path, 'static'] + json_file
        with open(join(*json_filepath_list)) as json_file:
            EMBEDDED_JAVASCRIPT += json_template_fmtstr % (json_file.read(),)

    embedded = dict(globals(), **locals())
    return appf.template('turk', 'identification_insert', **embedded)


@register_api('/api/status/query/graph/v2/', methods=['GET'], __api_plural_check__=False)
def view_graphs_status(ibs):
    graph_dict = {}
    for graph_uuid in current_app.GRAPH_CLIENT_DICT:
        graph_client = current_app.GRAPH_CLIENT_DICT.get(graph_uuid, None)
        if graph_client is None:
            continue
        graph_status, graph_exception = graph_client.refresh_status()
        if graph_client.review_dict is None:
            num_edges = None
        else:
            edge_list = list(graph_client.review_dict.keys())
            num_edges = len(edge_list)
        graph_uuid = str(graph_uuid)
        graph_dict[graph_uuid] = {
            'status': graph_status,
            'num_aids': len(graph_client.aids),
            'num_reviews': num_edges,
        }
    return graph_dict


@register_ibs_method
@register_api('/api/review/query/graph/v2/', methods=['POST'])
def process_graph_match_html_v2(ibs, graph_uuid, **kwargs):
    graph_client, _ = ibs.get_graph_client_query_chips_graph_v2(graph_uuid)
    response_tuple = process_graph_match_html(ibs, **kwargs)
    (
        annot_uuid_1,
        annot_uuid_2,
        decision,
        tags,
        user_id,
        confidence,
        user_times,
    ) = response_tuple
    aid1 = ibs.get_annot_aids_from_uuid(annot_uuid_1)
    aid2 = ibs.get_annot_aids_from_uuid(annot_uuid_2)
    edge = (
        aid1,
        aid2,
    )
    user_id = controller_inject.get_user()
    now = datetime.utcnow()

    if decision in ['excludetop', 'excludebottom']:
        aid = aid1 if decision == 'excludetop' else aid2

        metadata_dict = ibs.get_annot_metadata(aid)
        assert 'excluded' not in metadata_dict
        metadata_dict['excluded'] = True
        ibs.set_annot_metadata([aid], [metadata_dict])

        payload = {
            'action': 'remove_annots',
            'aids': [aid],
        }
    elif decision in ['excludeleft', 'excluderight']:
        aid = aid1 if decision == 'excludeleft' else aid2

        metadata_dict = ibs.get_annot_metadata(aid)
        assert 'excluded' not in metadata_dict
        metadata_dict['excluded'] = True
        ibs.set_annot_metadata([aid], [metadata_dict])

        payload = {
            'action': 'remove_annots',
            'aids': [aid],
        }
    else:
        payload = {
            'action': 'add_feedback',
            'edge': edge,
            'evidence_decision': decision,
            # TODO: meta_decision should come from the html resp.  When generating
            # the html page, the default value should be its previous value. If the
            # user changes it to be something incompatible them perhaps just reset
            # it to null.
            'meta_decision': 'null',
            'tags': [] if len(tags) == 0 else tags.split(';'),
            'user_id': 'user:web:%s' % (user_id,),
            'confidence': confidence,
            'timestamp_s1': user_times['server_time_start'],
            'timestamp_c1': user_times['client_time_start'],
            'timestamp_c2': user_times['client_time_end'],
            'timestamp': float(now.strftime('%s.%f')),
        }
    logger.info('POSTING GRAPH CLIENT REVIEW:')
    logger.info(ut.repr4(payload))
    graph_client.post(payload)

    # Clean any old continue_reviews
    graph_client.cleanup()

    # Continue review
    future = graph_client.post({'action': 'continue_review'})
    future.graph_client = graph_client
    future.add_done_callback(query_graph_v2_on_request_review)

    f2 = graph_client.post({'action': 'latest_logs'})
    f2.graph_client = graph_client
    f2.add_done_callback(query_graph_v2_latest_logs)
    return (
        annot_uuid_1,
        annot_uuid_2,
    )


@register_ibs_method
@register_api('/api/query/graph/v2/', methods=['GET'])
def sync_query_chips_graph_v2(ibs, graph_uuid):
    import wbia

    graph_client, _ = ibs.get_graph_client_query_chips_graph_v2(graph_uuid)

    # Create the AnnotInference
    infr = wbia.AnnotInference(ibs=ibs, aids=graph_client.aids, autoinit=True)
    for key in graph_client.config:
        infr.params[key] = graph_client.config[key]
    infr.reset_feedback('staging', apply=True)

    infr.relabel_using_reviews(rectify=True)
    edge_delta_df = infr.match_state_delta(old='annotmatch', new='all')
    name_delta_df = infr.get_wbia_name_delta()

    ############################################################################

    col_list = list(edge_delta_df.columns)
    match_aid_edge_list = list(edge_delta_df.index)
    match_aid1_list = ut.take_column(match_aid_edge_list, 0)
    match_aid2_list = ut.take_column(match_aid_edge_list, 1)
    match_annot_uuid1_list = ibs.get_annot_uuids(match_aid1_list)
    match_annot_uuid2_list = ibs.get_annot_uuids(match_aid2_list)
    match_annot_uuid_edge_list = list(zip(match_annot_uuid1_list, match_annot_uuid2_list))

    zipped = list(zip(*(list(edge_delta_df[col]) for col in col_list)))

    match_list = []
    for match_annot_uuid_edge, zipped_ in list(zip(match_annot_uuid_edge_list, zipped)):
        match_dict = {
            'edge': match_annot_uuid_edge,
        }
        for index, col in enumerate(col_list):
            match_dict[col] = zipped_[index]
        match_list.append(match_dict)

    ############################################################################

    col_list = list(name_delta_df.columns)
    name_aid_list = list(name_delta_df.index)
    name_annot_uuid_list = ibs.get_annot_uuids(name_aid_list)
    old_name_list = list(name_delta_df['old_name'])
    new_name_list = list(name_delta_df['new_name'])
    zipped = list(zip(name_annot_uuid_list, old_name_list, new_name_list))
    name_dict = {
        str(name_annot_uuid): {'old': old_name, 'new': new_name}
        for name_annot_uuid, old_name, new_name in zipped
    }

    ############################################################################

    ret_dict = {
        'match_list': match_list,
        'name_dict': name_dict,
    }

    infr.write_wbia_staging_feedback()
    infr.write_wbia_annotmatch_feedback(edge_delta_df)
    infr.write_wbia_name_assignment(name_delta_df)
    edge_delta_df.reset_index()

    return ret_dict


@register_ibs_method
@register_api('/api/query/graph/v2/', methods=['PUT'])
def add_annots_query_chips_graph_v2(ibs, graph_uuid, annot_uuid_list):
    graph_client, _ = ibs.get_graph_client_query_chips_graph_v2(graph_uuid)
    ibs.web_check_uuids(qannot_uuid_list=annot_uuid_list)
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)

    for graph_uuid_ in current_app.GRAPH_CLIENT_DICT:
        graph_client_ = current_app.GRAPH_CLIENT_DICT[graph_uuid_]
        aid_list_ = graph_client_.aids
        assert aid_list_ is not None
        overlap_aid_set = set(aid_list_) & set(aid_list)
        if len(overlap_aid_set) > 0:
            overlap_aid_list = list(overlap_aid_set)
            overlap_annot_uuid_list = ibs.get_annot_uuids(overlap_aid_list)
            raise controller_inject.WebUnavailableUUIDException(
                overlap_annot_uuid_list, graph_uuid_
            )

    aid_list_ = graph_client.aids + aid_list
    graph_uuid_ = ut.hashable_to_uuid(sorted(aid_list_))
    assert graph_uuid_ not in current_app.GRAPH_CLIENT_DICT
    graph_client.graph_uuid = graph_uuid_

    payload = {
        'action': 'add_annots',
        'dbdir': ibs.dbdir,
        'aids': aid_list,
    }
    future = graph_client.post(payload)
    future.result()  # Guarantee that this has happened before calling refresh

    # Start main loop
    future = graph_client.post({'action': 'continue_review'})
    future.graph_client = graph_client
    future.add_done_callback(query_graph_v2_on_request_review)

    current_app.GRAPH_CLIENT_DICT[graph_uuid_] = graph_client
    current_app.GRAPH_CLIENT_DICT[graph_uuid] = graph_uuid_
    return graph_uuid_


@register_ibs_method
def remove_annots_query_chips_graph_v2(ibs, graph_uuid, annot_uuid_list):
    graph_client, _ = ibs.get_graph_client_query_chips_graph_v2(graph_uuid)
    ibs.web_check_uuids(qannot_uuid_list=annot_uuid_list)
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)

    aid_list_ = list(set(graph_client.aids) - set(aid_list))
    graph_uuid_ = ut.hashable_to_uuid(sorted(aid_list_))
    assert graph_uuid_ not in current_app.GRAPH_CLIENT_DICT
    graph_client.graph_uuid = graph_uuid_

    payload = {
        'action': 'remove_annots',
        'dbdir': ibs.dbdir,
        'aids': aid_list,
    }
    future = graph_client.post(payload)
    future.result()  # Guarantee that this has happened before calling refresh

    # Start main loop
    future = graph_client.post({'action': 'continue_review'})
    future.graph_client = graph_client
    future.add_done_callback(query_graph_v2_on_request_review)

    current_app.GRAPH_CLIENT_DICT[graph_uuid_] = graph_client
    current_app.GRAPH_CLIENT_DICT[graph_uuid] = graph_uuid_
    return graph_uuid_


@register_ibs_method
@register_api('/api/query/graph/v2/', methods=['DELETE'])
def delete_query_chips_graph_v2(ibs, graph_uuid):
    values = ibs.get_graph_client_query_chips_graph_v2(graph_uuid)
    graph_client, graph_uuid_chain = values
    del graph_client
    for graph_uuid_ in graph_uuid_chain:
        if graph_uuid_ in current_app.GRAPH_CLIENT_DICT:
            current_app.GRAPH_CLIENT_DICT[graph_uuid_] = None
            current_app.GRAPH_CLIENT_DICT.pop(graph_uuid_)
    return True


def query_graph_v2_latest_logs(future):
    if not future.cancelled():
        logs = future.result()
        logger.info('--- <LOG DUMP> ---')
        for msg, color in logs:
            ut.cprint('[web.infr] ' + msg, color)
        logger.info('--- <\\LOG DUMP> ---')


def query_graph_v2_on_request_review(future):
    if not future.cancelled():
        graph_client = future.graph_client
        data_list = future.result()
        if data_list is not None:
            graph_client.update(data_list)
            callback_type = 'review'
        else:
            graph_client.update(None)
            callback_type = 'finished'
        query_graph_v2_callback(graph_client, callback_type)



# -*- coding: utf-8 -*-
"""
python -c "import utool as ut; ut.write_modscript_alias('Tgen.sh', 'wbia.templates.template_generator')"
sh Tgen.sh --key review --invert --Tcfg with_getters=True with_setters=False --modfname manual_review_funcs

# TODO: Fix this name it is too special case
"""
import logging
import six  # NOQA
from six.moves import zip, map, reduce

# import numpy as np
# import vtool as vt
import numpy as np
import ubelt as ub  # NOQA
from wbia import constants as const
from wbia.control import accessor_decors, controller_inject  # NOQA
import utool as ut
import uuid
from wbia.control.controller_inject import make_ibs_register_decorator

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


VERBOSE_SQL = ut.get_argflag(('--print-sql', '--verbose-sql', '--verb-sql', '--verbsql'))
CLASS_INJECT_KEY, register_ibs_method = make_ibs_register_decorator(__name__)


register_api = controller_inject.get_wbia_flask_api(__name__)


REVIEW_ROWID = 'review_rowid'
REVIEW_UUID = 'review_uuid'
REVIEW_AID1 = 'annot_1_rowid'
REVIEW_AID2 = 'annot_2_rowid'
REVIEW_COUNT = 'review_count'
REVIEW_EVIDENCE_DECISION = 'review_evidence_decision'
REVIEW_META_DECISION = 'review_meta_decision'
REVIEW_USER_IDENTITY = 'review_user_identity'
REVIEW_USER_CONFIDENCE = 'review_user_confidence'
REVIEW_TAGS = 'review_tags'
REVIEW_TIME_CLIENT_START = 'review_client_start_time_posix'
REVIEW_TIME_CLIENT_END = 'review_client_end_time_posix'
REVIEW_TIME_SERVER_START = 'review_server_start_time_posix'
REVIEW_TIME_SERVER_END = 'review_server_end_time_posix'


def e_(u, v):
    return (u, v) if u < v else (v, u)


def hack_create_aidpair_index(ibs):
    # HACK IN INDEX
    sqlfmt = ut.codeblock(
        """
        CREATE INDEX IF NOT EXISTS {index_name} ON {table} ({index_cols});
        """
    )
    sqlcmd = sqlfmt.format(
        index_name='aidpair_to_rowid',
        table=ibs.const.REVIEW_TABLE,
        index_cols=','.join([REVIEW_AID1, REVIEW_AID2]),
    )
    ibs.staging.connection.execute(sqlcmd).fetchall()
    sqlcmd = sqlfmt.format(
        index_name='aid1_to_rowids',
        table=ibs.const.REVIEW_TABLE,
        index_cols=','.join([REVIEW_AID1]),
    )
    ibs.staging.connection.execute(sqlcmd).fetchall()
    sqlcmd = sqlfmt.format(
        index_name='aid2_to_rowids',
        table=ibs.const.REVIEW_TABLE,
        index_cols=','.join([REVIEW_AID2]),
    )
    ibs.staging.connection.execute(sqlcmd).fetchall()


@register_ibs_method
@accessor_decors.ider
@register_api('/api/review/', methods=['GET'])
def _get_all_review_rowids(ibs):
    r"""
    Returns:
        list_ (list): all nids of known animals
        (does not include unknown names)
    """
    all_known_review_rowids = ibs.staging.get_all_rowids(const.REVIEW_TABLE)
    return all_known_review_rowids


@register_ibs_method
def _set_review_uuids(ibs, review_rowid_list, review_uuid_list):
    r"""
    Returns:
        list_ (list): all nids of known animals
        (does not include unknown names)
    """
    id_iter = ((review_rowid,) for review_rowid in review_rowid_list)
    val_iter = ((review_uuid,) for review_uuid in review_uuid_list)
    ibs.staging.set(const.REVIEW_TABLE, (REVIEW_UUID,), val_iter, id_iter)


@register_ibs_method
def get_review_rowid_from_superkey(
    ibs, aid_1_list, aid_2_list, count_list, eager=False, nInput=None
):
    """Returns review_rowid_list

    Args:
        superkey lists: review_rowid_list, aid_list

    Returns:
        review_rowid_list
    """
    colnames = (REVIEW_ROWID,)
    params_iter = zip(aid_1_list, aid_2_list, count_list)
    where_colnames = [REVIEW_AID1, REVIEW_AID2, REVIEW_COUNT]
    review_rowid_list = list(
        ibs.staging.get_where_eq(
            const.REVIEW_TABLE,
            colnames,
            params_iter,
            where_colnames,
            eager=eager,
            nInput=nInput,
        )
    )
    return review_rowid_list


@register_ibs_method
@accessor_decors.adder
@register_api('/api/review/', methods=['POST'])
def add_review(
    ibs,
    aid_1_list,
    aid_2_list,
    evidence_decision_list,
    meta_decision_list=None,
    review_uuid_list=None,
    identity_list=None,
    user_confidence_list=None,
    tags_list=None,
    review_client_start_time_posix=None,
    review_client_end_time_posix=None,
    review_server_start_time_posix=None,
    review_server_end_time_posix=None,
):
    r"""
    Adds a list of reviews.

    Returns:
        list: review_id_list - review rowids

    RESTful:
        Method: POST
        URL:    /api/review/

    CommandLine:
        python -m wbia.control.manual_review_funcs --test-add_review

    Doctest:
        >>> import wbia
        >>> from wbia.control.manual_review_funcs import *
        >>> ibs = wbia.opendb('testdb1')
        >>> ibs.staging.get_table_as_pandas('reviews')
        >>> # ensure it is empty
        >>> rowids = ibs.staging.get_all_rowids('reviews')
        >>> ibs.staging.delete_rowids('reviews', rowids)
        >>> ut.exec_funckw(ibs.add_review, globals())
        >>> # Add some dummy reviews
        >>> aid_1_list = [1, 2, 3, 2]
        >>> aid_2_list = [2, 3, 4, 3]
        >>> evidence_decision_list = [1, 0, 1, 2]
        >>> new_rowids = ibs.add_review(aid_1_list, aid_2_list,
        >>>                             evidence_decision_list)
        >>> assert new_rowids == [1, 2, 3, 4]
        >>> table = ibs.staging.get_table_as_pandas('reviews')
        >>> print(table)
        >>> # Then delete them
        >>> ibs.staging.delete_rowids('reviews', new_rowids)
    """
    assert len(aid_1_list) == len(aid_2_list)
    assert len(aid_1_list) == len(evidence_decision_list)
    diff_list = -np.array(aid_2_list)
    assert np.all(diff_list != 0), 'Cannot add a review state between an aid and itself'
    n_input = len(aid_1_list)

    # Order aid_1_list and aid_2_list pairs so that aid_1_list is always lower
    aid_pair_list = [e_(u, v) for u, v in zip(aid_1_list, aid_2_list)]
    aid_1_list = [pair[0] for pair in aid_pair_list]
    aid_2_list = [pair[1] for pair in aid_pair_list]

    if True:
        # Get current review counts from database
        unique_pairs = list(set(aid_pair_list))
        count_base = [
            0 if counts is None or len(counts) == 0 else max(max(counts), len(counts))
            for counts in ibs.get_review_counts_from_pairs(unique_pairs)
        ]
        pair_to_count = {edge: count for edge, count in zip(unique_pairs, count_base)}
        count_list = []
        for edge in aid_pair_list:
            pair_to_count[edge] += 1
            count = pair_to_count[edge]
            count_list.append(count)
    # else:
    #     # Alternative implemenation
    #     unique_pairs, groupxs = ut.group_indices(aid_pair_list)
    #     unique_base = [
    #         0 if counts is None or len(counts) == 0 else
    #         max(max(counts), len(counts))
    #         for counts in ibs.get_review_counts_from_pairs(unique_pairs)
    #     ]
    #     grouped_base = [[b] * len(g) for b, g in zip(unique_base, groupxs)]
    #     grouped_offsets = [list(range(n)) for n in map(len, groupxs)]
    #     base = np.array(ut.ungroup(grouped_base, groupxs))
    #     offsets = np.array(ut.ungroup(grouped_offsets, groupxs))
    #     count_list = offsets + base + 1

    if review_uuid_list is None:
        review_uuid_list = [uuid.uuid4() for _ in range(n_input)]
    if meta_decision_list is None:
        meta_decision_list = [None for _ in range(n_input)]
    if identity_list is None:
        # identity_list = [ut.get_computer_name()] * len(aid_1_list)
        identity_list = [None] * n_input
    if tags_list is None:
        tag_str_list = [None] * n_input
    else:
        tag_str_list = [';'.join(map(str, tag_list)) for tag_list in tags_list]
    if user_confidence_list is None:
        user_confidence_list = [None] * n_input

    if review_client_start_time_posix is None:
        review_client_start_time_posix = [None] * n_input
    if review_client_end_time_posix is None:
        review_client_end_time_posix = [None] * n_input
    if review_server_start_time_posix is None:
        review_server_start_time_posix = [None] * n_input
    if review_server_end_time_posix is None:
        review_server_end_time_posix = [None] * n_input

    assert n_input == len(identity_list)
    assert n_input == len(tag_str_list)
    assert n_input == len(user_confidence_list)
    assert n_input == len(review_uuid_list)
    assert n_input == len(count_list)

    superkey_paramx = (
        0,
        1,
        2,
    )
    # TODO Allow for better ensure=False without using partial
    # Just autogenerate these functions
    colnames = [
        REVIEW_AID1,
        REVIEW_AID2,
        REVIEW_COUNT,
        REVIEW_UUID,
        REVIEW_EVIDENCE_DECISION,
        REVIEW_META_DECISION,
        REVIEW_USER_IDENTITY,
        REVIEW_USER_CONFIDENCE,
        REVIEW_TAGS,
        REVIEW_TIME_CLIENT_START,
        REVIEW_TIME_CLIENT_END,
        REVIEW_TIME_SERVER_START,
        REVIEW_TIME_SERVER_END,
    ]
    params_iter = list(
        zip(
            aid_1_list,
            aid_2_list,
            count_list,
            review_uuid_list,
            evidence_decision_list,
            meta_decision_list,
            identity_list,
            user_confidence_list,
            tag_str_list,
            review_client_start_time_posix,
            review_client_end_time_posix,
            review_server_start_time_posix,
            review_server_end_time_posix,
        )
    )
    review_rowid_list = ibs.staging.add_cleanly(
        const.REVIEW_TABLE,
        colnames,
        params_iter,
        ibs.get_review_rowid_from_superkey,
        superkey_paramx,
    )
    return review_rowid_list


@register_ibs_method
@accessor_decors.deleter
# @cache_invalidator(const.REVIEW_TABLE)
@register_api('/api/review/', methods=['DELETE'])
def delete_review(ibs, review_rowid_list):
    r"""
    deletes reviews from the database

    RESTful:
        Method: DELETE
        URL:    /api/review/
    """
    if ut.VERBOSE:
        logger.info('[ibs] deleting %d reviews' % len(review_rowid_list))
    ibs.staging.delete_rowids(const.REVIEW_TABLE, review_rowid_list)


@register_ibs_method
def get_review_rowids_from_edges(ibs, edges, eager=True, nInput=None, directed=False):
    colnames = (REVIEW_ROWID,)
    # Order aid_1_list and aid_2_list pairs so that aid_1_list is always lower
    # params_iter = edges
    if directed:
        params_iter = edges
    else:
        params_iter = [e_(u, v) for u, v in edges]
    where_colnames = [REVIEW_AID1, REVIEW_AID2]
    review_rowids_list = ibs.staging.get_where_eq(
        const.REVIEW_TABLE,
        colnames,
        params_iter,
        where_colnames,
        eager=eager,
        nInput=nInput,
        unpack_scalars=False,
    )
    return review_rowids_list


@register_ibs_method
@accessor_decors.getter_1to1
def get_review_exists_from_edges(ibs, edges, eager=True, nInput=None):
    # Order aid_1_list and aid_2_list pairs so that aid_1_list is always lower
    # params_iter = (e_(u, v) for u, v in edges)
    params_iter = edges
    where_colnames = [REVIEW_AID1, REVIEW_AID2]
    exists_list = ibs.staging.exists_where_eq(
        const.REVIEW_TABLE,
        params_iter,
        where_colnames,
        eager=False,
        nInput=nInput,
        unpack_scalars=True,
    )
    exists_list = map(bool, exists_list)
    if eager:
        exists_list = list(exists_list)
    return exists_list


@register_ibs_method
@accessor_decors.getter_1toM
@register_api('/api/review/rowids/tuple/', methods=['GET'], __api_plural_check__=False)
def get_review_rowids_from_aid_tuple(
    ibs, aid_1_list, aid_2_list, eager=True, nInput=None
):
    r"""
    Aid pairs are undirected

    Returns:
        list_ (list): review_rowid_list - review rowid list of lists

    RESTful:
        Method: GET
        URL:    /api/review/rowid/tuple/
    """
    # Order aid_1_list and aid_2_list pairs so that aid_1_list is always lower
    edges = (e_(u, v) for u, v in zip(aid_1_list, aid_2_list))
    return get_review_rowids_from_edges(ibs, edges, eager=eager, nInput=nInput)
    # colnames = (REVIEW_ROWID,)
    # where_colnames = [REVIEW_AID1, REVIEW_AID2]
    # review_rowids_list = ibs.staging.get_where_eq(
    #     const.REVIEW_TABLE, colnames, params_iter, where_colnames,
    #     eager=eager, nInput=nInput, unpack_scalars=False)
    # return review_rowids_list


@register_ibs_method
def get_review_rowids_between(ibs, aids1, aids2=None, method=None):
    """
    Find staging rowids between sets of aids

    Doctest:
        >>> from wbia.control.manual_review_funcs import *
        >>> import wbia
        >>> ibs = wbia.opendb('PZ_MTEST')
        >>> aids1 = aids2 = [1, 2, 3, 4, 5, 6]
        >>> rowids_between = ibs.get_review_rowids_between
        >>> ids1 = sorted(rowids_between(aids1, aids2, method=1))
        >>> ids2 = sorted(rowids_between(aids1, aids2, method=2))
        >>> assert len(ub.find_duplicates(ids1)) == 0
        >>> assert len(ub.find_duplicates(ids2)) == 0
        >>> assert ids1 == ids2
    """
    if aids2 is None:
        aids2 = aids1
    if method is None:
        if len(aids1) * len(aids2) > 5000:
            method = 1
        else:
            method = 2

    if method == 1:
        # Strategy 1: get all existing rows and see what intersects
        # This is better when the enumerated set of rows would be larger than
        # the database size
        rowids11 = set(ut.flatten(ibs.get_review_rowids_from_aid1(aids1)))
        rowids12 = set(ut.flatten(ibs.get_review_rowids_from_aid2(aids1)))
        if aids1 is aids2:
            rowids = list(reduce(set.intersection, [rowids11, rowids12]))
        else:
            rowids21 = set(ut.flatten(ibs.get_review_rowids_from_aid1(aids2)))
            rowids22 = set(ut.flatten(ibs.get_review_rowids_from_aid2(aids2)))
            rowids = list(
                reduce(set.intersection, [rowids11, rowids12, rowids21, rowids22])
            )
    elif method == 2:
        # Strategy 2: enumerate what rows could exist and see what does exist
        # This is better when the enumerated set of rows would be smaller than
        # the database size
        edges = list(ut.product_nonsame(aids1, aids2))
        if len(edges) == 0:
            rowids = []
        else:
            rowids = ibs.get_review_rowids_from_edges(edges, directed=True)
            if rowids is None:
                rowids = []
            rowids = ut.filter_Nones(rowids)
            rowids = ut.flatten(rowids)
    else:
        raise ValueError('no method=%r' % (method,))
    return rowids


# @register_ibs_method
# @accessor_decors.getter_1to1
# @register_api('/api/review/rowids/undirected/', methods=['GET'], __api_plural_check__=False)
# def get_review_rowids_from_undirected_tuple(ibs, aid_1_list, aid_2_list):
#     aids1, aids2 = aid_1_list, aid_2_list
#     review_rowids_dir1 = ibs.get_review_rowids_from_aid_tuple(aids1, aids2)
#     review_rowids_dir2 = ibs.get_review_rowids_from_aid_tuple(aids2, aids1)
#     def _join_rowids(rowids1, rowids2):
#         if rowids1 is None and rowids2 is None:
#             return None
#         else:
#             if rowids1 is None:
#                 rowids1 = []
#             elif rowids2 is None:
#                 rowids2 = []
#             return rowids1 + rowids2
#     review_rowids_list = [
#         _join_rowids(rowids1, rowids2)
#         for rowids1, rowids2 in zip(review_rowids_dir1, review_rowids_dir2)
#     ]
#     return review_rowids_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/review/count/', methods=['GET'])
def get_review_count(ibs, review_rowid_list):
    review_count_list = ibs.staging.get(
        const.REVIEW_TABLE, (REVIEW_COUNT,), review_rowid_list
    )
    return review_count_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/review/counts/tuple/', methods=['GET'], __api_plural_check__=False)
def get_review_counts_from_tuple(ibs, aid_1_list, aid_2_list, eager=True, nInput=None):
    r"""
    Returns:
        list_ (list): review_counts_list - review counts

    RESTful:
        Method: GET
        URL:    /api/review/counts/tuple/
    """
    aid_pairs = zip(aid_1_list, aid_2_list)
    review_counts_list = ibs.get_review_counts_from_pairs(aid_pairs)
    return review_counts_list


@register_ibs_method
def get_review_counts_from_pairs(ibs, aid_pairs, eager=True, nInput=None):
    r"""
    Returns:
        list_ (list): review_counts_list - review counts

    RESTful:
        Method: GET
        URL:    /api/review/counts/tuple/
    """
    colnames = (REVIEW_COUNT,)
    params_iter = aid_pairs
    where_colnames = [REVIEW_AID1, REVIEW_AID2]
    review_counts_list = ibs.staging.get_where_eq(
        const.REVIEW_TABLE,
        colnames,
        params_iter,
        where_colnames,
        eager=eager,
        nInput=nInput,
        unpack_scalars=False,
    )
    return review_counts_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/review/decision/', methods=['GET'])
def get_review_decision(ibs, review_rowid_list):
    review_decision_list = ibs.staging.get(
        const.REVIEW_TABLE, (REVIEW_EVIDENCE_DECISION,), review_rowid_list
    )
    return review_decision_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/review/uuid/', methods=['GET'])
def get_review_uuid(ibs, review_rowid_list):
    review_uuid_list = ibs.staging.get(
        const.REVIEW_TABLE, (REVIEW_UUID,), review_rowid_list
    )
    return review_uuid_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/review/decision/str/', methods=['GET'])
def get_review_decision_str(ibs, review_rowid_list):
    review_decision_list = ibs.get_review_decision(review_rowid_list)
    review_decision_str_list = [
        const.EVIDENCE_DECISION.INT_TO_NICE.get(review_decision)
        for review_decision in review_decision_list
    ]
    return review_decision_str_list


@register_ibs_method
@register_api('/api/review/decisions/only/', methods=['GET'], __api_plural_check__=False)
def get_review_decisions_from_only(ibs, aid_list, eager=True, nInput=None):
    r"""
    Returns:
        list_ (list): review_tuple_decisions_list - review decisions

    RESTful:
        Method: GET
        URL:    /api/review/decisions/only/
    """
    colnames = (
        REVIEW_AID1,
        REVIEW_AID2,
        REVIEW_EVIDENCE_DECISION,
    )
    params_iter = [(aid,) for aid in aid_list]
    where_clause = '%s=?' % (REVIEW_AID1)
    review_tuple_decisions_list = ibs.staging.get_where(
        const.REVIEW_TABLE, colnames, params_iter, where_clause, unpack_scalars=False
    )
    return review_tuple_decisions_list


@register_ibs_method
@register_api('/api/review/rowids/only/', methods=['GET'], __api_plural_check__=False)
def get_review_rowids_from_only(ibs, aid_list, eager=True, nInput=None):
    r"""
    Returns:
        list_ (list): review_rowids

    RESTful:
        Method: GET
        URL:    /api/review/rowids/only/
    """
    colnames = (REVIEW_ROWID,)
    params_iter = [(aid,) for aid in aid_list]
    where_clause = '%s=?' % (REVIEW_AID1)
    review_rowids = ibs.staging.get_where(
        const.REVIEW_TABLE, colnames, params_iter, where_clause, unpack_scalars=False
    )
    return review_rowids


@register_ibs_method
def get_review_rowids_from_single(ibs, aid_list, eager=True, nInput=None):
    colnames = (REVIEW_ROWID,)
    params_iter = [(aid, aid) for aid in aid_list]
    where_clause = '%s=? OR %s=?' % (REVIEW_AID1, REVIEW_AID2)
    review_rowids = ibs.staging.get_where(
        const.REVIEW_TABLE,
        colnames,
        params_iter,
        where_clause=where_clause,
        unpack_scalars=False,
    )
    return review_rowids


@register_ibs_method
def get_review_rowids_from_aid1(ibs, aid_list, eager=True, nInput=None):
    colnames = (REVIEW_ROWID,)
    params_iter = [(aid,) for aid in aid_list]
    where_clause = '%s=?' % (REVIEW_AID1,)
    review_rowids = ibs.staging.get_where(
        const.REVIEW_TABLE,
        colnames,
        params_iter,
        where_clause=where_clause,
        unpack_scalars=False,
    )
    return review_rowids


@register_ibs_method
def get_review_rowids_from_aid2(ibs, aid_list, eager=True, nInput=None):
    colnames = (REVIEW_ROWID,)
    params_iter = [(aid,) for aid in aid_list]
    where_clause = '%s=?' % (REVIEW_AID2,)
    review_rowids = ibs.staging.get_where(
        const.REVIEW_TABLE,
        colnames,
        params_iter,
        where_clause=where_clause,
        unpack_scalars=False,
    )
    return review_rowids


# @register_ibs_method
# @accessor_decors.getter_1to1
# @register_api('/api/review/decisions/single/', methods=['GET'], __api_plural_check__=False)
# def get_review_decisions_from_single(ibs, aid_list, eager=True, nInput=None):
#     r"""
#     Returns:
#         list_ (list): review_tuple_decisions_list - review decisions

#     RESTful:
#         Method: GET
#         URL:    /api/review/identities/single/
#     """
#     colnames = (REVIEW_AID1, REVIEW_AID2, REVIEW_EVIDENCE_DECISION,)
#     params_iter = zip(aid_list, aid_list, )
#     where_colnames = [REVIEW_AID1, REVIEW_AID2]
#     review_tuple_decisions_list = ibs.staging.get_where_eq(
#         const.REVIEW_TABLE, colnames, params_iter, where_colnames,
#         eager=eager, nInput=nInput, op='OR', unpack_scalars=False)
#     return review_tuple_decisions_list


# @register_ibs_method
# @accessor_decors.getter_1to1
# @register_api('/api/review/decisions/tuple/', methods=['GET'], __api_plural_check__=False)
# def get_review_decisions_from_tuple(ibs, aid_1_list, aid_2_list, eager=True, nInput=None):
#     r"""
#     Returns:
#         list_ (list): review_decisions_list - review decisions

#     RESTful:
#         Method: GET
#         URL:    /api/review/identities/tuple/
#     """
#     colnames = (REVIEW_EVIDENCE_DECISION,)
#     params_iter = zip(aid_1_list, aid_2_list)
#     where_colnames = [REVIEW_AID1, REVIEW_AID2]
#     review_decisions_list = ibs.staging.get_where_eq(
#         const.REVIEW_TABLE, colnames, params_iter, where_colnames,
#         eager=eager, nInput=nInput, unpack_scalars=False)
#     return review_decisions_list


# @register_ibs_method
# @accessor_decors.getter_1to1
# @register_api('/api/review/decisions/str/tuple/', methods=['GET'], __api_plural_check__=False)
# def get_review_decisions_str_from_tuple(ibs, aid_1_list, aid_2_list, **kwargs):
#     r"""
#     Returns:
#         list_ (list): review_decisions_list - review decisions

#     RESTful:
#         Method: GET
#         URL:    /api/review/identities/str/tuple/
#     """
#     review_decisions_list = ibs.get_review_decisions_from_tuple(
#         aid_1_list, aid_2_list, **kwargs)
#     review_decision_str_list = [
#         [
#             const.EVIDENCE_DECISION.INT_TO_NICE.get(review_decision)
#             for review_decision in review_decision_list
#         ]
#         for review_decision_list in review_decisions_list
#     ]
#     return review_decision_str_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/review/identity/', methods=['GET'])
def get_review_identity(ibs, review_rowid_list):
    review_identity_list = ibs.staging.get(
        const.REVIEW_TABLE, (REVIEW_USER_IDENTITY,), review_rowid_list
    )
    return review_identity_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/review/confidence/', methods=['GET'])
def get_review_user_confidence(ibs, review_rowid_list):
    user_confidence_list = ibs.staging.get(
        const.REVIEW_TABLE, (REVIEW_USER_CONFIDENCE,), review_rowid_list
    )
    return user_confidence_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api(
    '/api/review/identities/tuple/', methods=['GET'], __api_plural_check__=False
)
def get_review_identities_from_tuple(
    ibs, aid_1_list, aid_2_list, eager=True, nInput=None
):
    r"""
    Returns:
        list_ (list): review_identities_list - review identities

    RESTful:
        Method: GET
        URL:    /api/review/identities/tuple/
    """
    colnames = (REVIEW_USER_IDENTITY,)
    params_iter = zip(aid_1_list, aid_2_list)
    where_colnames = [REVIEW_AID1, REVIEW_AID2]
    review_identities_list = ibs.staging.get_where_eq(
        const.REVIEW_TABLE,
        colnames,
        params_iter,
        where_colnames,
        eager=eager,
        nInput=nInput,
        unpack_scalars=False,
    )
    return review_identities_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/review/time/posix/', methods=['GET'])
def get_review_posix_time(ibs, review_rowid_list):
    return ibs.get_review_posix_server_end_time(review_rowid_list)


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/review/time/server/start/posix/', methods=['GET'])
def get_review_posix_server_start_time(ibs, review_rowid_list):
    review_posix_time_list = ibs.staging.get(
        const.REVIEW_TABLE, (REVIEW_TIME_SERVER_START,), review_rowid_list
    )
    return review_posix_time_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/review/time/client/start/posix/', methods=['GET'])
def get_review_posix_client_start_time(ibs, review_rowid_list):
    review_posix_time_list = ibs.staging.get(
        const.REVIEW_TABLE, (REVIEW_TIME_CLIENT_START,), review_rowid_list
    )
    return review_posix_time_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/review/time/client/end/posix/', methods=['GET'])
def get_review_posix_client_end_time(ibs, review_rowid_list):
    review_posix_time_list = ibs.staging.get(
        const.REVIEW_TABLE, (REVIEW_TIME_CLIENT_END,), review_rowid_list
    )
    return review_posix_time_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/review/time/server/end/posix/', methods=['GET'])
def get_review_posix_server_end_time(ibs, review_rowid_list):
    review_posix_time_list = ibs.staging.get(
        const.REVIEW_TABLE, (REVIEW_TIME_SERVER_END,), review_rowid_list
    )
    return review_posix_time_list


@register_ibs_method
@accessor_decors.getter_1to1
def get_review_aid_tuple(ibs, review_rowid_list, eager=True, nInput=None):
    colnames = (
        REVIEW_AID1,
        REVIEW_AID2,
    )
    params_iter = zip(review_rowid_list)
    where_colnames = [REVIEW_ROWID]
    aid_tuple_list = ibs.staging.get_where_eq(
        const.REVIEW_TABLE,
        colnames,
        params_iter,
        where_colnames,
        eager=eager,
        nInput=nInput,
    )
    return aid_tuple_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/review/time/posix/tuple/', methods=['GET'])
def get_review_posix_times_from_tuple(
    ibs, aid_1_list, aid_2_list, eager=True, nInput=None
):
    r"""
    Returns:
        list_ (list): identity_list - review posix times

    RESTful:
        Method: GET
        URL:    /api/review/time/posix/tuple/
    """
    colnames = (REVIEW_TIME_SERVER_END,)
    params_iter = zip(aid_1_list, aid_2_list)
    where_colnames = [REVIEW_AID1, REVIEW_AID2]
    review_posix_times_list = ibs.staging.get_where_eq(
        const.REVIEW_TABLE,
        colnames,
        params_iter,
        where_colnames,
        eager=eager,
        nInput=nInput,
        unpack_scalars=False,
    )
    return review_posix_times_list


# def _parse_tag_str(tag_str):
#     if tag_str is None or len(tag_str) == 0:
#         return None
#     else:
#         return tag_str.split(';')


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/review/tags/', methods=['GET'], __api_plural_check__=False)
def get_review_tags(ibs, review_rowid_list):
    review_tag_str_list = ibs.staging.get(
        const.REVIEW_TABLE, (REVIEW_TAGS,), review_rowid_list
    )
    review_tags_list = [
        None
        if review_tag_str is None or len(review_tag_str) == 0
        else review_tag_str.split(';')
        for review_tag_str in review_tag_str_list
    ]
    return review_tags_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/review/tags/tuple/', methods=['GET'], __api_plural_check__=False)
def get_review_tags_from_tuple(ibs, aid_1_list, aid_2_list, eager=True, nInput=None):
    r"""
    Returns:
        list_ (list): review_tags_list - review tags (list of strings)

    RESTful:
        Method: GET
        URL:    /api/review/tags/tuple/
    """
    colnames = (REVIEW_TAGS,)
    params_iter = zip(aid_1_list, aid_2_list)
    where_colnames = [REVIEW_AID1, REVIEW_AID2]
    review_tag_strs_list = ibs.staging.get_where_eq(
        const.REVIEW_TABLE,
        colnames,
        params_iter,
        where_colnames,
        eager=eager,
        nInput=nInput,
        unpack_scalars=False,
    )
    review_tags_list = [
        []
        if review_tag_str_list is None
        else [
            None
            if review_tag_str is None or len(review_tag_str) == 0
            else review_tag_str.split(';')
            for review_tag_str in review_tag_str_list
        ]
        for review_tag_str_list in review_tag_strs_list
    ]
    return review_tags_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/review/metadata/', methods=['GET'])
def get_review_metadata(ibs, review_rowid_list, return_raw=False):
    r"""
    Returns:
        list_ (list): review metadata dictionary

    RESTful:
        Method: GET
        URL:    /api/review/metadata/
    """
    metadata_str_list = ibs.staging.get(
        const.REVIEW_TABLE, ('review_metadata_json',), review_rowid_list
    )
    metadata_list = []
    for metadata_str in metadata_str_list:
        if metadata_str in [None, '']:
            metadata_dict = {}
        else:
            metadata_dict = metadata_str if return_raw else ut.from_json(metadata_str)
        metadata_list.append(metadata_dict)
    return metadata_list


@register_ibs_method
@accessor_decors.setter
@register_api('/api/review/metadata/', methods=['PUT'])
def set_review_metadata(ibs, review_rowid_list, metadata_dict_list):
    r"""
    Sets the review's metadata using a metadata dictionary

    RESTful:
        Method: PUT
        URL:    /api/review/metadata/

    CommandLine:
        python -m wbia.control.manual_review_funcs --test-set_review_metadata

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_review_funcs import *  # NOQA
        >>> import wbia
        >>> import random
        >>> # build test data
        >>> ibs = wbia.opendb('testdb1')
        >>> review_rowid_list = ibs.add_review([1], [2], [0])
        >>> metadata_dict_list = [
        >>>     {'test': random.uniform(0.0, 1.0)},
        >>> ]
        >>> print(ut.repr2(metadata_dict_list))
        >>> ibs.set_review_metadata(review_rowid_list, metadata_dict_list)
        >>> # verify results
        >>> metadata_dict_list_ = ibs.get_review_metadata(review_rowid_list)
        >>> print(ut.repr2(metadata_dict_list_))
        >>> assert metadata_dict_list == metadata_dict_list_
        >>> metadata_str_list = [ut.to_json(metadata_dict) for metadata_dict in metadata_dict_list]
        >>> print(ut.repr2(metadata_str_list))
        >>> metadata_str_list_ = ibs.get_review_metadata(review_rowid_list, return_raw=True)
        >>> print(ut.repr2(metadata_str_list_))
        >>> assert metadata_str_list == metadata_str_list_
        >>> ibs.delete_review(review_rowid_list)
    """
    id_iter = ((review_rowid,) for review_rowid in review_rowid_list)
    metadata_str_list = []
    for metadata_dict in metadata_dict_list:
        metadata_str = ut.to_json(metadata_dict)
        metadata_str_list.append(metadata_str)
    val_list = ((metadata_str,) for metadata_str in metadata_str_list)
    ibs.staging.set(const.REVIEW_TABLE, ('review_metadata_json',), val_list, id_iter)



# -*- coding: utf-8 -*-
import logging
from vtool._pyflann_backend import pyflann as pyflann
import utool as ut
import uuid
import six
from os.path import exists, join
import lockfile

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')


class Win32CompatTempFile(object):
    """
    mimics tempfile.NamedTemporaryFile but allows the file to be closed without
    being deleted.  This lets a second process (like the FLANN) read/write to
    the file in a win32 system. The file is instead deleted after the
    Win32CompatTempFile object goes out of scope.

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.smk.pickle_flann import *  # NOQA
        >>> verbose = True
        >>> temp = Win32CompatTempFile(verbose=verbose)
        >>> data = six.b(str('10010'))
        >>> print('data = %r' % (data,))
        >>> data1 = temp.read()
        >>> print('data1 = %r' % (data1,))
        >>> temp.write(data)
        >>> data2 = temp.read()
        >>> print('data2 = %r' % (data2,))
        >>> temp.close()
        >>> assert data != data1
        >>> assert data == data2
        >>> ut.assert_raises(ValueError, temp.close)
        >>> assert not ut.checkpath(temp.fpath, verbose=verbose)
    """

    def __init__(temp, delete=True, verbose=False):
        temp.delete = delete
        appname = 'wbia'
        temp.dpath = ut.ensure_app_resource_dir(appname, 'tempfiles')
        temp.fpath = None
        temp.fname = None
        temp._isclosed = False
        temp.verbose = verbose
        temp._create_unique_file()

    @property
    def name(temp):
        return temp.fpath

    def read(temp):
        temp._check_open()
        with open(temp.fpath, 'rb') as file_:
            return file_.read()

    def write(temp, data):
        temp._check_open()
        with open(temp.fpath, 'wb') as file_:
            file_.write(data)
            file_.flush()

    def close(temp):
        temp._check_open()
        if temp.delete and exists(temp.fpath):
            ut.delete(temp.fpath, verbose=temp.verbose)
        temp._isclosed = True

    def _create_unique_file(temp):
        temp._check_open()
        with lockfile.LockFile(join(temp.dpath, 'tempfile.lock')):
            flag = True
            while flag or exists(temp.fpath):
                temp.fname = six.text_type(uuid.uuid4()) + '.temp'
                temp.fpath = join(temp.dpath, temp.fname)
                flag = False
            ut.touch(temp.fpath, verbose=temp.verbose)

    def _check_open(temp):
        if temp._isclosed:
            raise ValueError('I/O operation on closed object')

    def __del__(temp):
        if not temp._isclosed:
            temp.close()


if pyflann is not None:

    class PickleFLANN(pyflann.FLANN):
        """
        Adds the ability to pickle a flann class on a unix system.
        (Actually, pickle still wont work because we need the original point data.
        But we can do a custom dumps and a loads)

        CommandLine:
            python -m wbia.algo.smk.pickle_flann PickleFLANN

        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.algo.smk.pickle_flann import *  # NOQA
            >>> import numpy as np
            >>> rng = np.random.RandomState(42)
            >>> data = rng.rand(10, 2)
            >>> query = rng.rand(5, 2)
            >>> flann = PickleFLANN()
            >>> flann.build_index(data, random_seed=42)
            >>> index_bytes = flann.dumps()
            >>> flann2 = PickleFLANN()
            >>> flann2.loads(index_bytes, data)
            >>> assert flann2 is not flann
            >>> assert flann2.dumps() == index_bytes
            >>> idx1 = flann.nn_index(query)[0]
            >>> idx2 = flann2.nn_index(query)[0]
            >>> assert np.all(idx1 == idx2)
        """

        def dumps(self):
            """
            # Make a special wordflann pickle
            http://www.linuxscrew.com/2010/03/24/fastest-way-to-create-ramdisk-in-ubuntulinux/
            sudo mkdir /tmp/ramdisk; chmod 777 /tmp/ramdisk
            sudo mount -t tmpfs -o size=256M tmpfs /tmp/ramdisk/
            http://zeblog.co/?p=1588
            """
            # import tempfile
            # assert not ut.WIN32, 'Fix on WIN32. Cannot write to temp file'
            # temp = tempfile.NamedTemporaryFile(delete=True)
            temp = Win32CompatTempFile(delete=True, verbose=False)
            try:
                self.save_index(temp.name)
                index_bytes = temp.read()
            except Exception:
                raise
            finally:
                temp.close()
            return index_bytes

        def loads(self, index_bytes, pts):
            # import tempfile
            # assert not ut.WIN32, 'Fix on WIN32. Cannot write to temp file'
            # temp = tempfile.NamedTemporaryFile(delete=True)
            temp = Win32CompatTempFile(delete=True, verbose=False)
            try:
                temp.write(index_bytes)
                # temp.file.flush()
                self.load_index(temp.name, pts)
            except Exception:
                raise
            finally:
                temp.close()


else:
    PickleFLANN = None



# -*- coding: utf-8 -*-
import logging
from os.path import join, dirname, split, basename, splitext
import re
import utool as ut
from six.moves import map, range

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


class IndividualResultsCopyTaskQueue(object):
    def __init__(self):
        self.cp_task_list = []

    def append_copy_task(self, fpath_orig, dstdir=None):
        """ helper which copies a summary figure to root dir """
        fname_orig, ext = splitext(basename(fpath_orig))
        outdir = dirname(fpath_orig)
        fdir_clean, cfgdir = split(outdir)
        if dstdir is None:
            dstdir = fdir_clean
        # aug = cfgdir[0:min(len(cfgdir), 10)]
        aug = cfgdir
        fname_fmt = '{aug}_{fname_orig}{ext}'
        fmt_dict = {'aug': aug, 'fname_orig': fname_orig, 'ext': ext}
        fname_clean = ut.long_fname_format(
            fname_fmt, fmt_dict, ['fname_orig'], max_len=128
        )
        fdst_clean = join(dstdir, fname_clean)
        self.cp_task_list.append((fpath_orig, fdst_clean))

    def flush_copy_tasks(self):
        # Execute all copy tasks and empty the lists
        if ut.NOT_QUIET:
            logger.info('[DRAW_RESULT] copying %r summaries' % (len(self.cp_task_list)))
        for src, dst in self.cp_task_list:
            ut.copy(src, dst, verbose=False)
        del self.cp_task_list[:]


def make_individual_latex_figures(
    ibs, fpaths_list, flat_case_labels, cfgx2_shortlbl, case_figdir, analysis_fpath_list
):
    # HACK MAKE LATEX CONVINENCE STUFF
    # logger.info('LATEX HACK')
    if len(fpaths_list) == 0:
        logger.info('nothing to render')
        return
    RENDER = ut.get_argflag('--render')
    DUMP_FIGDEF = ut.get_argflag(('--figdump', '--dump-figdef', '--figdef'))

    if not (DUMP_FIGDEF or RENDER):  # HACK
        return

    latex_code_blocks = []
    latex_block_keys = []

    caption_prefix = ut.get_argval('--cappref', type_=str, default='')
    caption_suffix = ut.get_argval('--capsuf', type_=str, default='')
    cmdaug = ut.get_argval('--cmdaug', type_=str, default='custom')

    selected = None

    for case_idx, (fpaths, labels) in enumerate(zip(fpaths_list, flat_case_labels)):
        if labels is None:
            labels = [cmdaug]
        if len(fpaths) < 4:
            nCols = len(fpaths)
        else:
            nCols = 2

        _cmdname = ibs.get_dbname() + ' Case ' + ' '.join(labels) + '_' + str(case_idx)
        # logger.info('_cmdname = %r' % (_cmdname,))
        cmdname = ut.latex_sanitize_command_name(_cmdname)
        label_str = cmdname
        if len(caption_prefix) == 0:
            caption_str = ut.escape_latex(
                'Casetags: '
                + ut.repr2(labels, nl=False, strvals=True)
                + ', db='
                + ibs.get_dbname()
                + '. '
            )
        else:
            caption_str = ''

        use_sublbls = len(cfgx2_shortlbl) > 1
        if use_sublbls:
            caption_str += ut.escape_latex(
                'Each figure shows a different configuration: '
            )
            sublbls = [
                '(' + chr(97 + count) + ') ' for count in range(len(cfgx2_shortlbl))
            ]
        else:
            # caption_str += ut.escape_latex('This figure depicts correct and
            # incorrect matches from configuration: ')
            sublbls = [''] * len(cfgx2_shortlbl)

        def wrap_tt(text):
            return r'{\tt ' + text + '}'

        _shortlbls = cfgx2_shortlbl
        _shortlbls = list(map(ut.escape_latex, _shortlbls))
        # Adjust spacing for breaks
        # tex_small_space = r''
        tex_small_space = r'\hspace{0pt}'
        # Remove query specific config flags in individual results
        _shortlbls = [re.sub('\\bq[^,]*,?', '', shortlbl) for shortlbl in _shortlbls]
        # Let config strings be broken over newlines
        _shortlbls = [
            re.sub('\\+', tex_small_space + '+' + tex_small_space, shortlbl)
            for shortlbl in _shortlbls
        ]
        _shortlbls = [
            re.sub(', *', ',' + tex_small_space, shortlbl) for shortlbl in _shortlbls
        ]
        _shortlbls = list(map(wrap_tt, _shortlbls))
        cfgx2_texshortlbl = [
            '\n    ' + lbl + shortlbl for lbl, shortlbl in zip(sublbls, _shortlbls)
        ]

        caption_str += ut.conj_phrase(cfgx2_texshortlbl, 'and') + '.\n    '
        caption_str = '\n    ' + caption_prefix + caption_str + caption_suffix
        caption_str = caption_str.rstrip()
        figure_str = ut.get_latex_figure_str(
            fpaths,
            nCols=nCols,
            label_str=label_str,
            caption_str=caption_str,
            use_sublbls=None,
            use_frame=True,
        )
        latex_block = ut.latex_newcommand(cmdname, figure_str)
        latex_block = '\n%----------\n' + latex_block
        latex_code_blocks.append(latex_block)
        latex_block_keys.append(cmdname)

    # HACK
    remove_fpath = ut.truepath('~/latex/crall-candidacy-2015') + '/'

    latex_fpath = join(case_figdir, 'latex_cases.tex')

    if selected is not None:
        selected_keys = selected
    else:
        selected_keys = latex_block_keys

    selected_blocks = ut.dict_take(
        dict(zip(latex_block_keys, latex_code_blocks)), selected_keys
    )

    figdef_block = '\n'.join(selected_blocks)
    figcmd_block = '\n'.join(['\\' + key for key in latex_block_keys])

    selected_block = figdef_block + '\n\n' + figcmd_block

    # HACK: need full paths to render
    selected_block_renderable = selected_block
    selected_block = selected_block.replace(remove_fpath, '')
    if RENDER:
        ut.render_latex_text(selected_block_renderable)

    if DUMP_FIGDEF:
        ut.writeto(latex_fpath, selected_block)

    # if NOT DUMP AND NOT RENDER:
    #    logger.info('STANDARD LATEX RESULTS')
    #    cmdname = ibs.get_dbname() + 'Results'
    #    latex_block  = ut.get_latex_figure_str2(analysis_fpath_list, cmdname, nCols=1)
    #    ut.print_code(latex_block, 'latex')
    if DUMP_FIGDEF or RENDER:
        ut.print_code(selected_block, 'latex')



# -*- coding: utf-8 -*-
"""
Dependencies: flask, tornado

SeeAlso:
    routes.turk_identification
"""
import logging
from wbia.control import controller_inject
from flask import url_for, request, current_app  # NOQA
import numpy as np  # NOQA
import utool as ut
import uuid
import requests
import six
import json

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')

CLASS_INJECT_KEY, register_ibs_method = controller_inject.make_ibs_register_decorator(
    __name__
)
register_api = controller_inject.get_wbia_flask_api(__name__)


REMOTE_TESTING = True


if not REMOTE_TESTING:
    REMOTE_DOMAIN = '127.0.0.1'
    REMOTE_PORT = '5555'
    REMOTE_UUID = None
else:
    # info for flukebook background server
    REMOTE_DOMAIN = 'kaiju.dyn.wildme.io'
    REMOTE_PORT = '5015'
    # https://kaiju.dyn.wildme.io:5015/api/core/db/uuid/init/
    REMOTE_UUID = '511ab1d0-5af4-4808-b0a7-c8115240ab8e'

    remote_args = ut.get_arg_dict()
    REMOTE_DOMAIN = remote_args.get('sync-domain', REMOTE_DOMAIN)
    REMOTE_PORT = remote_args.get('sync-port', REMOTE_PORT)
    REMOTE_UUID = remote_args.get('sync-uuid', REMOTE_UUID)

    if REMOTE_UUID in [True, '', 'none', 'None']:
        REMOTE_UUID = None


REMOTE_URL = 'http://%s:%s' % (REMOTE_DOMAIN, REMOTE_PORT)
REMOTE_UUID = None if REMOTE_UUID is None else uuid.UUID(REMOTE_UUID)


def _construct_route_url(route_rule):
    if not route_rule.startswith('/'):
        route_rule = '/' + route_rule
    if not route_rule.endswith('/'):
        route_rule = route_rule + '/'
    route_url = '%s%s' % (REMOTE_URL, route_rule)
    return route_url


@register_ibs_method
def _construct_route_url_ibs(ibs, route_rule):
    return _construct_route_url(route_rule)


def _verify_response(response):
    try:
        response_dict = ut.from_json(response.text)
    except ValueError:
        raise AssertionError('Could not get valid JSON response from server')
    status = response_dict.get('status', {})
    assert status.get('success', False)
    response = response_dict.get('response', None)
    return response


@register_ibs_method
def _verify_response_ibs(ibs, response):
    try:
        response_dict = ut.from_json(response.text)
    except ValueError:
        raise AssertionError('Could not get valid JSON response from server')
    status = response_dict.get('status', {})
    assert status.get('success', False)
    response = response_dict.get('response', None)
    return response


def _get(route_rule, **kwargs):
    route_url = _construct_route_url(route_rule)
    response = requests.get(route_url, **kwargs)
    return _verify_response(response)


@register_ibs_method
def _get_ibs(ibs, route_rule, **kwargs):
    route_url = ibs._construct_route_url_ibs(route_rule)
    response = requests.get(route_url, **kwargs)
    return ibs._verify_response_ibs(response)


def _assert_remote_online(ibs):
    try:
        version = _get('/api/core/db/version/')
        uuid = _get('/api/core/db/uuid/init/')
        assert version == ibs.get_database_version()
        if REMOTE_UUID is not None:
            assert uuid == REMOTE_UUID
    except Exception:
        raise IOError('Remote IBEIS DETECT database offline at %s' % (REMOTE_URL,))


@register_ibs_method
def _detect_remote_push_images(ibs, gid_list):
    route_url = _construct_route_url('/api/upload/image/')

    num_images = len(gid_list)
    image_path_list = ibs.get_image_paths(gid_list)
    for index, image_path in enumerate(image_path_list):
        logger.info('\tSending %d / %d: %r' % (index, num_images, image_path))
        file_dict = {
            'image': open(image_path, 'rb'),
        }
        response = requests.post(route_url, files=file_dict)
        _verify_response(response)
        logger.info('\t...sent')


@register_ibs_method
def _detect_remote_push_imageset(ibs, image_uuid_list):
    route_url = _construct_route_url('/api/image/imageset/text/json/')

    db_name = ibs.get_dbname()
    db_uuid = ibs.get_db_init_uuid()
    time_str = ut.get_timestamp()
    imageset_text = 'Sync from %s (%s) at %s' % (db_name, db_uuid, time_str)
    imageset_text_list = [imageset_text] * len(image_uuid_list)

    data_dict = {
        'image_uuid_list': image_uuid_list,
        'imageset_text_list': imageset_text_list,
    }
    for key in data_dict:
        data_dict[key] = ut.to_json(data_dict[key])
    response = requests.put(route_url, data=data_dict)
    _verify_response(response)


@register_ibs_method
def _detect_remote_push_annots(ibs, aid_list):
    route_url = _construct_route_url('/api/annot/json/')

    logger.info('\tSending...')
    data_dict = {
        'image_uuid_list': ibs.get_annot_image_uuids(aid_list),
        'annot_uuid_list': ibs.get_annot_uuids(aid_list),
        'annot_bbox_list': ibs.get_annot_bboxes(aid_list),
    }
    for key in data_dict:
        data_dict[key] = ut.to_json(data_dict[key])
    response = requests.post(route_url, data=data_dict)
    _verify_response(response)
    logger.info('\t...sent')


@register_ibs_method
def _detect_remote_push_metadata(
    ibs, route_rule, uuid_str, value_str, uuid_list, value_list
):
    route_url = _construct_route_url(route_rule)

    logger.info('\tSetting %s metadata for %s' % (route_rule, uuid_str))
    data_dict = {
        uuid_str: uuid_list,
        value_str: value_list,
    }
    for key in data_dict:
        data_dict[key] = ut.to_json(data_dict[key])
    response = requests.put(route_url, data=data_dict)
    _verify_response(response)
    logger.info('\t...set')


@register_ibs_method
def _detect_remote_push_annot_metadata(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    ibs._detect_remote_push_metadata(
        '/api/annot/bbox/json/',
        'annot_uuid_list',
        'bbox_list',
        annot_uuid_list,
        ibs.get_annot_bboxes(aid_list),
    )
    ibs._detect_remote_push_metadata(
        '/api/annot/theta/json/',
        'annot_uuid_list',
        'theta_list',
        annot_uuid_list,
        ibs.get_annot_thetas(aid_list),
    )
    ibs._detect_remote_push_metadata(
        '/api/annot/viewpoint/json/',
        'annot_uuid_list',
        'viewpoint_list',
        annot_uuid_list,
        ibs.get_annot_viewpoints(aid_list),
    )
    ibs._detect_remote_push_metadata(
        '/api/annot/quality/text/json/',
        'annot_uuid_list',
        'quality_text_list',
        annot_uuid_list,
        ibs.get_annot_quality_texts(aid_list),
    )
    ibs._detect_remote_push_metadata(
        '/api/annot/species/json/',
        'annot_uuid_list',
        'species_text_list',
        annot_uuid_list,
        ibs.get_annot_species_texts(aid_list),
    )
    ibs._detect_remote_push_metadata(
        '/api/annot/multiple/json/',
        'annot_uuid_list',
        'flag_list',
        annot_uuid_list,
        ibs.get_annot_multiple(aid_list),
    )
    ibs._detect_remote_push_metadata(
        '/api/annot/interest/json/',
        'annot_uuid_list',
        'flag_list',
        annot_uuid_list,
        ibs.get_annot_interest(aid_list),
    )
    ibs._detect_remote_push_metadata(
        '/api/annot/tags/json/',
        'annot_uuid_list',
        'annot_tags_list',
        annot_uuid_list,
        ibs.get_annot_tag_text(aid_list),
    )
    ibs._detect_remote_push_metadata(
        '/api/annot/name/text/json/',
        'annot_uuid_list',
        'name_text_list',
        annot_uuid_list,
        ibs.get_annot_name_texts(aid_list),
    )


@register_ibs_method
def _detect_remote_push_parts(ibs, part_rowid_list):
    route_url = _construct_route_url('/api/part/json/')

    logger.info('\tSending...')
    data_dict = {
        'annot_uuid_list': ibs.get_part_annot_uuids(part_rowid_list),
        'part_uuid_list': ibs.get_part_uuids(part_rowid_list),
        'part_bbox_list': ibs.get_part_bboxes(part_rowid_list),
    }
    for key in data_dict:
        data_dict[key] = ut.to_json(data_dict[key])
    response = requests.post(route_url, data=data_dict)
    _verify_response(response)
    logger.info('\t...sent')


@register_ibs_method
def _detect_remote_push_part_metadata(ibs, part_uuid_list):
    part_rowid_list = ibs.get_part_rowids_from_uuid(part_uuid_list)
    ibs._detect_remote_push_metadata(
        '/api/part/bbox/json/',
        'part_uuid_list',
        'bbox_list',
        part_uuid_list,
        ibs.get_part_bboxes(part_rowid_list),
    )
    ibs._detect_remote_push_metadata(
        '/api/part/theta/json/',
        'part_uuid_list',
        'theta_list',
        part_uuid_list,
        ibs.get_part_thetas(part_rowid_list),
    )
    ibs._detect_remote_push_metadata(
        '/api/part/viewpoint/json/',
        'part_uuid_list',
        'viewpoint_list',
        part_uuid_list,
        ibs.get_part_viewpoints(part_rowid_list),
    )
    ibs._detect_remote_push_metadata(
        '/api/part/quality/text/json/',
        'part_uuid_list',
        'quality_text_list',
        part_uuid_list,
        ibs.get_part_quality_texts(part_rowid_list),
    )  # NOQA
    ibs._detect_remote_push_metadata(
        '/api/part/type/json/',
        'part_uuid_list',
        'type_text_list',
        part_uuid_list,
        ibs.get_part_types(part_rowid_list),
    )
    ibs._detect_remote_push_metadata(
        '/api/part/tags/json/',
        'part_uuid_list',
        'part_tags_list',
        part_uuid_list,
        ibs.get_part_tag_text(part_rowid_list),
    )


@register_ibs_method
def _sync_get_remote_info(ibs):
    route_url = _construct_route_url('/api/core/db/info/')
    response = requests.get(route_url)
    _verify_response(response)
    logger.info(response.text)
    return response


@register_ibs_method
def _sync_get_species_aids(ibs, species_name):
    route_url = _construct_route_url('/api/annot/')
    data_dict = {
        'species': species_name,
    }
    response = requests.get(route_url, data=data_dict)
    _verify_response(response)
    return response.json()['response']


# eubalaena_australis, eubalaena_glacialis
@register_ibs_method
def _sync_get_image(ibs, remote_image_rowid):
    endpoint = '/api/image/%s/' % remote_image_rowid
    return _get(endpoint)


@register_ibs_method
def sync_get_training_data(ibs, species_name, force_update=False, **kwargs):

    aid_list = ibs._sync_get_training_aids(species_name, **kwargs)
    ann_uuids = ibs._sync_get_annot_endpoint('/api/annot/uuid/', aid_list)
    local_aids = []

    # avoid re-downloading annots based on UUID
    if not force_update:
        local_ann_uuids = set(ibs.get_valid_annot_uuids())
        remote_ann_uuids = set(ann_uuids)
        dupe_uuids = local_ann_uuids & remote_ann_uuids  # & is set-union
        new_uuids = remote_ann_uuids - dupe_uuids
        local_aids = ibs.get_annot_aids_from_uuid(dupe_uuids)
        aid_list = ibs._sync_get_aids_for_uuids(new_uuids)
        ann_uuids = ibs._sync_get_annot_endpoint('/api/annot/uuid/', aid_list)

    # get needed info
    viewpoints = ibs._sync_get_annot_endpoint('/api/annot/viewpoint/', aid_list)
    bboxes = ibs._sync_get_annot_endpoint('/api/annot/bbox/', aid_list)
    thetas = ibs._sync_get_annot_endpoint('/api/annot/theta/', aid_list)
    name_texts = ibs._sync_get_annot_endpoint('/api/annot/name/text/', aid_list)
    name_uuids = ibs._sync_get_annot_endpoint('/api/annot/name/uuid/', aid_list)
    images = ibs._sync_get_annot_endpoint('/api/annot/image/rowid/', aid_list)
    gpaths = [ibs._construct_route_url_ibs('/api/image/src/%s/' % gid) for gid in images]
    specieses = [species_name] * len(aid_list)

    gid_list = ibs.add_images(gpaths)
    nid_list = ibs.add_names(name_texts, name_uuids)

    local_aids += ibs.add_annots(
        gid_list,
        bbox_list=bboxes,
        theta_list=thetas,
        species_list=specieses,
        nid_list=nid_list,
        annot_uuid_list=ann_uuids,
        viewpoint_list=viewpoints,
    )

    return local_aids


@register_ibs_method
def sync_get_training_data_uuid_list(ibs, auuid_list, force_update=False, **kwargs):

    aid_list = ibs._sync_get_auuid_endpoint('/api/annot/rowid/uuid/', auuid_list)
    ann_uuids = ibs._sync_get_annot_endpoint('/api/annot/uuid/', aid_list)
    local_aids = []

    # avoid re-downloading annots based on UUID
    if not force_update:
        local_ann_uuids = set(ibs.get_valid_annot_uuids())
        remote_ann_uuids = set(ann_uuids)
        dupe_uuids = local_ann_uuids & remote_ann_uuids  # & is set-union
        new_uuids = remote_ann_uuids - dupe_uuids
        local_aids = ibs.get_annot_aids_from_uuid(dupe_uuids)
        aid_list = ibs._sync_get_aids_for_uuids(new_uuids)
        ann_uuids = ibs._sync_get_annot_endpoint('/api/annot/uuid/', aid_list)

    # get needed info
    viewpoints = ibs._sync_get_annot_endpoint('/api/annot/viewpoint/', aid_list)
    bboxes = ibs._sync_get_annot_endpoint('/api/annot/bbox/', aid_list)
    thetas = ibs._sync_get_annot_endpoint('/api/annot/theta/', aid_list)
    name_texts = ibs._sync_get_annot_endpoint('/api/annot/name/text/', aid_list)
    name_uuids = ibs._sync_get_annot_endpoint('/api/annot/name/uuid/', aid_list)
    images = ibs._sync_get_annot_endpoint('/api/annot/image/rowid/', aid_list)
    gpaths = [ibs._construct_route_url_ibs('/api/image/src/%s/' % gid) for gid in images]
    specieses = ibs._sync_get_annot_endpoint('/api/annot/species/', aid_list)

    gid_list = ibs.add_images(gpaths)
    nid_list = ibs.add_names(name_texts, name_uuids)

    local_aids += ibs.add_annots(
        gid_list,
        bbox_list=bboxes,
        theta_list=thetas,
        species_list=specieses,
        nid_list=nid_list,
        annot_uuid_list=ann_uuids,
        viewpoint_list=viewpoints,
    )

    return local_aids


@register_ibs_method
def _sync_get_names(ibs, aid_list):
    return ibs._sync_get_annot_endpoint('/api/annot/name/rowid/', aid_list)


# @register_ibs_method
# def _sync_get_remote_name_uuids(ibs, local_nids):
#     return ibs._sync_get_annot_endpoint('/api/annot/name/rowid/', aid_list)


@register_ibs_method
def _sync_get_remote_name_uuids(ibs, local_nids):
    return ibs._sync_get_annot_endpoint('/api/annot/name/rowid/', local_nids)


@register_ibs_method
def _sync_get_annot_endpoint(ibs, endpoint, aid_list):
    route_url = _construct_route_url(endpoint)
    logger.info('\tGetting info on %d aids from %s' % (len(aid_list), route_url))
    data_dict = {
        'aid_list': json.dumps(aid_list),
    }
    return _get(endpoint, data=data_dict)


@register_ibs_method
def _sync_get_auuid_endpoint(ibs, endpoint, auuid_list):

    # make sure auuid_list is correct format
    working_list = auuid_list.copy()
    if len(working_list) > 0 and type(working_list[0]) is str:
        from uuid import UUID

        working_list = [UUID(item) for item in working_list]

    serialized_uuids = ut.to_json(working_list)

    route_url = _construct_route_url(endpoint)
    logger.info('\tGetting info on %d auuids from %s' % (len(auuid_list), route_url))
    data_dict = {'uuid_list': serialized_uuids}
    return _get(endpoint, data=data_dict)


@register_ibs_method
def _sync_get_aids_for_uuids(ibs, uuids):
    data_dict = {
        'uuid_list': ut.to_json(list(uuids)),
    }
    return ibs._get_ibs('/api/annot/rowid/uuid/', data=data_dict)


@register_ibs_method
def _sync_filter_only_multiple_sightings(ibs, aid_list):
    r"""
    Returns:
        filtered_aids (list): the subset of aid_list such that every annot
        has a name and each name appears at least 2x.
    """
    name_list = ibs._sync_get_names(aid_list)
    name_hist = ut.dict_hist(name_list)
    aid_names = zip(aid_list, name_list)
    filtered_aids = [aid for (aid, name) in aid_names if name_hist[name] > 1]
    filtered_aid_names = [name for (aid, name) in aid_names if name_hist[name] > 1]
    return filtered_aids, filtered_aid_names


@register_ibs_method
def _sync_get_training_aids(ibs, species_name, limit=100000):
    aid_list = ibs._sync_get_species_aids(species_name)
    aid_list, name_list = ibs._sync_filter_only_multiple_sightings(aid_list)
    # if limit is not None:
    #     aid_list = ibs._sync_get_training_subset(aid_list, name_list, limit)
    return aid_list[:limit]


# @register_ibs_method
# def _sync_get_training_subset(ibs, aid_list, name_list, limit):
#     min_aid_per_name = 2
#     max_aid_per_name = 5
#     name_to_aids = {}
#     pass


@register_ibs_method
@register_api('/api/sync/', methods=['GET'])
def detect_remote_sync_images(ibs, gid_list=None, only_sync_missing_images=True):
    _assert_remote_online(ibs)

    if gid_list is None:
        gid_list = ibs.get_valid_gids()

    confirm_list = [ut.random_nonce()[:5] for _ in range(3)]
    confirm_str = '-'.join(confirm_list)
    logger.info(
        'You are about to submit %d images to a remote DETECT database at %r with UUID=%r.'
        % (len(gid_list), REMOTE_URL, REMOTE_UUID)
    )
    logger.info(
        'Only do this action if you are confident in the detection accuracy of the images, annotations, annotation metadata, parts and part metadata.'
    )
    logger.info(
        'In order to continue, please type exactly the confirmation string %r'
        % (confirm_str,)
    )

    if six.PY2:
        input_func = raw_input  # NOQA
    else:
        input_func = input
    response_str = input_func('Confirmation string [Empty to abort]: ')
    response_str = response_str.lower()
    assert confirm_str == response_str, 'Confirmation string mismatch, aborting...'

    ############################################################################

    # Sync images
    image_uuid_list = ibs.get_image_uuids(gid_list)
    image_uuid_list_ = _get('/api/image/json/')

    missing_gid_list = [
        gid
        for gid, image_uuid in list(zip(gid_list, image_uuid_list))
        if image_uuid not in image_uuid_list_
    ]

    num_missing = len(missing_gid_list)
    if num_missing > 0:
        logger.info('Need to push %d images...' % (num_missing,))
        ibs._detect_remote_push_images(missing_gid_list)
        logger.info('...pushed')

    # Filter only missing
    gid_list_ = missing_gid_list if only_sync_missing_images else gid_list
    image_uuid_list_ = ibs.get_image_uuids(gid_list_)

    ############################################################################

    # Sync imageset
    logger.info('Setting imageset...')
    ibs._detect_remote_push_imageset(image_uuid_list_)
    logger.info('...set')

    ############################################################################

    # Sync annots
    aid_list = ut.flatten(ibs.get_image_aids(gid_list_))
    annot_uuid_list = ibs.get_annot_uuids(aid_list)
    annot_uuid_list_ = _get('/api/annot/json/')

    missing_aid_list = [
        aid
        for aid, annot_uuid in list(zip(aid_list, annot_uuid_list))
        if annot_uuid not in annot_uuid_list_
    ]

    num_missing = len(missing_aid_list)
    if num_missing > 0:
        logger.info('Need to push %d annots...' % (num_missing,))
        ibs._detect_remote_push_annots(missing_aid_list)
        logger.info('...pushed')

    ############################################################################

    # Sync annotation metadata
    logger.info('Synching annotation metadata...')
    if len(annot_uuid_list) > 0:
        ibs._detect_remote_push_annot_metadata(annot_uuid_list)
    logger.info('...synched')

    ############################################################################

    # Sync parts
    part_rowid_list = ut.flatten(ibs.get_annot_part_rowids(aid_list))
    part_uuid_list = ibs.get_part_uuids(part_rowid_list)
    part_uuid_list_ = _get('/api/part/json/')

    missing_part_rowid_list = [
        part_rowid
        for part_rowid, part_uuid in list(zip(part_rowid_list, part_uuid_list))
        if part_uuid not in part_uuid_list_
    ]

    num_missing = len(missing_part_rowid_list)
    if num_missing > 0:
        logger.info('Need to push %d parts...' % (num_missing,))
        ibs._detect_remote_push_parts(missing_part_rowid_list)
        logger.info('...pushed')

    ############################################################################

    # Sync part metadata
    logger.info('Synching part metadata...')
    if len(part_uuid_list) > 0:
        ibs._detect_remote_push_part_metadata(part_uuid_list)
    logger.info('...synched')



# -*- coding: utf-8 -*-
from wbia.guitool.__PYQT__ import QtCore, QtWidgets

# from wbia.guitool import guitool_components
# (print, print_, printDBG, rrr, profile) = utool.inject(__name__, '[APIButtonWidget]', DEBUG=False)
import utool as ut

ut.noinject(__name__, '[api_timestamp_delegate]', DEBUG=False)


DELEGATE_BASE = QtWidgets.QItemDelegate
# DELEGATE_BASE = QtWidgets.QStyledItemDelegate


class APITimestampDelegate(DELEGATE_BASE):
    def __init__(dgt, parent=None):
        assert parent is not None, 'parent must be a view'
        DELEGATE_BASE.__init__(dgt, parent)

    def paint(dgt, painter, option, qtindex):
        painter.save()
        data = qtindex.model().data(qtindex, QtCore.Qt.DisplayRole)
        print(data)

        painter.restore()

    # def editorEvent(dgt, event, model, option, qtindex):
    #    event_type = event.type()
    #    if event_type == QtCore.QEvent.MouseButtonPress:
    #        # store the position that is clicked
    #        dgt._pressed = (qtindex.row(), qtindex.column())
    #        if utool.VERBOSE:
    #            print('store')
    #        return True
    #    elif event_type == QtCore.QEvent.MouseButtonRelease:
    #        if dgt.is_qtindex_pressed(qtindex):
    #            print('emit')
    #            dgt.button_clicked.emit(qtindex)
    #            pass
    #        elif dgt._pressed is not None:
    #            # different place.
    #            # force a repaint on the pressed cell by emitting a dataChanged
    #            # Note: This is probably not the best idea
    #            # but I've yet to find a better solution.
    #            print('repaint')
    #            oldIndex = qtindex.model().index(*dgt._pressed)
    #            dgt._pressed = None
    #            qtindex.model().dataChanged.emit(oldIndex, oldIndex)
    #            pass
    #        dgt._pressed = None
    #        #print('mouse release')
    #        return True
    #    elif event_type == QtCore.QEvent.Leave:
    #        print('leave')
    #    elif event_type == QtCore.QEvent.MouseButtonDblClick:
    #        print('doubleclick')
    #    else:
    #        print('event_type = %r' % event_type)
    #        return DELEGATE_BASE.editorEvent(dgt, event, model, option, qtindex)



# -*- coding: utf-8 -*-
"""Integrates numpy types into sqlite3"""
import io
import uuid
from sqlite3 import register_adapter, register_converter

import numpy as np
import utool as ut


__all__ = ()


def _read_numpy_from_sqlite3(blob):
    # INVESTIGATE: Is memory freed up correctly here?
    out = io.BytesIO(blob)
    out.seek(0)
    # return np.load(out)
    # Is this better?
    arr = np.load(out)
    out.close()
    return arr


def _read_bool(b):
    return None if b is None else bool(b)


def _write_bool(b):
    return b


def _write_numpy_to_sqlite3(arr):
    out = io.BytesIO()
    np.save(out, arr)
    out.seek(0)
    return memoryview(out.read())


def _read_uuid_from_sqlite3(blob):
    try:
        return uuid.UUID(bytes_le=blob)
    except ValueError as ex:
        ut.printex(ex, keys=['blob'])
        raise


def _read_dict_from_sqlite3(blob):
    return ut.from_json(blob)
    # return uuid.UUID(bytes_le=blob)


def _write_dict_to_sqlite3(dict_):
    return ut.to_json(dict_)


def _write_uuid_to_sqlite3(uuid_):
    return memoryview(uuid_.bytes_le)


def register_numpy_dtypes():
    py_int_type = int
    for dtype in (
        np.int8,
        np.int16,
        np.int32,
        np.int64,
        np.uint8,
        np.uint16,
        np.uint32,
        np.uint64,
    ):
        register_adapter(dtype, py_int_type)
    register_adapter(np.float32, float)
    register_adapter(np.float64, float)


def register_numpy():
    """
    Tell SQL how to deal with numpy arrays
    Utility function allowing numpy arrays to be stored as raw blob data
    """
    register_converter('NUMPY', _read_numpy_from_sqlite3)
    register_converter('NDARRAY', _read_numpy_from_sqlite3)
    register_adapter(np.ndarray, _write_numpy_to_sqlite3)


def register_uuid():
    """ Utility function allowing uuids to be stored in sqlite """
    register_converter('UUID', _read_uuid_from_sqlite3)
    register_adapter(uuid.UUID, _write_uuid_to_sqlite3)


def register_dict():
    register_converter('DICT', _read_dict_from_sqlite3)
    register_adapter(dict, _write_dict_to_sqlite3)


def register_list():
    register_converter('LIST', ut.from_json)
    register_adapter(list, ut.to_json)


# def register_bool():
#     # FIXME: ensure this works
#     register_converter('BOOL', _read_bool)
#     register_adapter(bool, _write_bool)


register_numpy_dtypes()
register_numpy()
register_uuid()
register_dict()
register_list()
# register_bool()  # TODO
